#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, struct, threading, time
from . rmcp import RMCP_Ping, RMCP
from . _rmcpp_msg import IPMI20_Message
from . _crypto import RAKP_NONE, MD5_128, get_cipher_tuple, \
                        cal_auth_code, conv_str2bytes
from . _rakp import *
from .. mesg.ipmi_app import IPMI_SendMsg, GetChnlAuthCap, SetSessPriv, CloseSess
from .. util.exception import PyIntfExcept, PyIntfSeqExcept

class RMCPP(RMCP):
    def __init__(self, opts, keep_alive):
        super(RMCPP, self).__init__(opts, keep_alive)
        self.cipher = (RAKP_NONE, RAKP_NONE, RAKP_NONE)
        self.k1 = None
        self.k2 = None
        self.sid = b'\0'

    def open(self, opts):
        cipher_suite = opts.get('cipher_suite', 3)        
        cipher = get_cipher_tuple(cipher_suite)
        if cipher is None:
            raise PyIntfExcept('Cipher suite {0} is not supported.'.format(cipher))

        # Open socket
        RMCP_Ping.open(self)

        user = conv_str2bytes(opts.get('user', None))
        self.passwd = conv_str2bytes(opts.get('password', None))
        if self.passwd is not None: self.passwd = struct.pack('20s', self.passwd)
        priv = opts.get('priv', 4)        
        self.rcsid = os.urandom(4)

        no_ping = opts.get('no_ping', False)
        if not no_ping:        
            # rmcp ping
            self.ping()

        # Get Channel Authentication Capabilities Command
        self.issue_cmd(GetChnlAuthCap, priv)

        # Open Session Request
        mssid, cipher = self.issue_cmd(RMCPP_OpenSess, self.rcsid, cipher, priv)

        # RAKP 1, 2
        rcrn = os.urandom(16)
        msrn, msguid = self.issue_cmd(RAKP_1_2, self.rcsid, mssid, rcrn, 
                                      priv, user, self.passwd, cipher[0])

        # Calculate SIK        
        # SIK = H(rcrn, msrn, priv, len(user), user)
        # key = KG => user password
        if 'kg' not in opts.keys() or opts['kg'] == '':
            key = self.passwd
        else:
            key = conv_str2bytes(opts['kg'])
            key = struct.pack('20s', key)
        
        if key is not None:
            data = rcrn + msrn
            if user is None:
                data += struct.pack('BB', priv, 0)
            else:
                data += struct.pack('BB', priv, len(user))
                data += user
        
            sik = cal_auth_code(cipher[0], key, data)
        else:
            sik = None

        # RAKP 3, 4
        self.issue_cmd(RAKP_3_4, 
                       self.rcsid, mssid, msrn, rcrn, msguid,
                       priv, user, self.passwd, cipher[0], sik)
        
        self.cipher = cipher
        self.cipher_suite = cipher_suite
        self.sess_act = True
        self.sid = mssid

        # Calculate K1 & K2 by need
        if self.cipher[1] != RAKP_NONE and self.cipher[1] != MD5_128:
            self.k1 = cal_auth_code(cipher[0], sik, b'\x01' * 20)
        if self.cipher[2] != RAKP_NONE:
            self.k2 = cal_auth_code(cipher[0], sik, b'\x02' * 20)

        # Set Session Privilege Level Command
        priv, = self.issue_cmd(SetSessPriv, priv)

        # Create the keep-alive thread
        if self.keep_alive:
            th = threading.Thread(target=self.keep_alive_cb)
            th.daemon = True
            th.start()

    def gen_msg(self, cmd, bridging=False, dest=0, target=0):
        msg = IPMI20_Message(self.cipher, self.sseq, self.sid, self.passwd, 
                             self.k1, self.k2)
                             
        if cmd.payload_type != 0:
            # RMCP Open Session Request & RAKP 1, 3
            payload = cmd.pack()
            data = msg.pack(cmd, payload)
            ret = (msg, data)
        elif bridging:
            # Message bridging
            inner = msg._pack_lan_payload(cmd, target)
            sm = IPMI_SendMsg(dest, inner)
            data = msg.pack(sm)
            ret = (msg, data, sm)
        else:
            # Common IPMI commands
            data = msg.pack(cmd)
            ret = (msg, data)

        return ret

    def unpack(self, rsp, msg, cmd):
        if cmd.payload_type != 0:
            payload_type, pkt1 = msg.unpack(rsp)  
            pkt2 = cmd.unpack(payload_type, pkt1)
        else:
            pkt1 = msg.unpack(rsp)  
            pkt2 = cmd.unpack(pkt1)

        return pkt2



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import builtins
from .. util.exception import PyExcept, PyCmdsExcept, PyCmdsArgsExcept

def help_raw(self, from_shell=True):
    print('Issue raw commands:')
    print('    raw <NetFn> <CMD> <Req Data>')

    if from_shell:
        print('    [<NetFn> <CMD> <Req Data>]')
        print('    [<NetFn>, <CMD>, <Req_b1, Req_b2, ...>]')

def _do_raw_imp(self, raw, from_shell):
    # At least 2 arguments for NetFn and CMD
    # unless it's asking for help
    if raw[0] == 'help':
        help_raw(self, from_shell)
        return
    elif len(raw) < 2:
        raise PyCmdsArgsExcept(1)

    # convert arguments to int from str
    req = []
    for i in raw:  # Drop the program name
        try:
            if not i: continue
            if i[:2] == '0x' or i[:2] == '0X':
                req.append(int(i, base=16))
            else:
                req.append(int(i))
        except:
            raise PyCmdsArgsExcept(3, 0, i) 

    rsp = self.intf.issue_raw_cmd(req)
    self.print_rsp(rsp)

def do_raw(self, text):
    from_shell = True

    # process input to a list
    if type(text) is str:
        # arguments separated by , or space
        raw = text.split(sep=',')
        if len(raw) < 2:
            raw = text.split(sep=' ')
    else:
        from_shell = False
        raw = text

    raw = [x.strip() for x in raw]

    try:
        _do_raw_imp(self, raw, from_shell)
        
    except PyCmdsExcept as e:
        builtins.print(e, '\n')
        if e.context >= 0:  help_raw(self, from_shell)

    except PyExcept as e:
        builtins.print(e)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import struct
from . import IPMI_Message
from .. util.exception import PyMesgExcept

NETFN_SE = 4

class GetSensorReading(IPMI_Message):
    def __init__(self, sensor_num):
        super(GetSensorReading, self).__init__(NETFN_SE, 0x2d)
        self.req_data = struct.pack('B', sensor_num)

    def unpack(self, rsp):
        cc, ret = rsp[2:]        
        if cc != 0:
            return super(GetSensorReading, self).unpack(rsp)
        if ret is None:
            raise PyMesgExcept('Get Sensor Reading: unexpected empty response data.') 

        rsp_data_len = len(ret)
        if  rsp_data_len >= 2 or rsp_data_len <= 4:
            return super(GetSensorReading, self).unpack(rsp, 'B' * rsp_data_len)
        raise PyMesgExcept('Get Sensor Reading: unexpected response data len = {0}'.format(len(ret)))      

class GetSensorHys(IPMI_Message):
    def __init__(self, sensor_num):
        super(GetSensorHys, self).__init__(NETFN_SE, 0x25)
        self.req_data = struct.pack('BB', sensor_num, 0xff)

    def unpack(self, rsp):
        return super(GetSensorHys, self).unpack(rsp, 'BB')

class GetSensorThres(IPMI_Message):
    def __init__(self, sensor_num):
        super(GetSensorThres, self).__init__(NETFN_SE, 0x27)
        self.req_data = struct.pack('B', sensor_num)

    def unpack(self, rsp):
        return super(GetSensorThres, self).unpack(rsp, 'B' * 7)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from .. util.exception import PyIntfExcept

__all__ = [
    'kcs',
    'rmcp',
    'rmcpp',
    'ioctl',
    'init',    
]

def init(opts, ping_only, keep_alive):
    if ping_only:
        from . rmcp import RMCP_Ping
        intf = RMCP_Ping(opts['global'])
        intf.open()
        return intf

    if 'interface' not in opts['global']:
        intf_name = 'lanplus'
    else:
        intf_name = opts['global']['interface']

    if intf_name == 'lan':
        from . rmcp import RMCP
        intf = RMCP(opts['global'], keep_alive)
        intf.open(opts['lan'])
        return intf

    if intf_name == 'lanplus':
        from . rmcpp import RMCPP
        intf = RMCPP(opts['global'], keep_alive)
        intf.open(opts['lanplus'])
        return intf

    if intf_name == 'kcs':
        from . kcs import KCS
        intf = KCS(opts['kcs'])
        intf.open()
        return intf

    raise PyIntfExcept('Invalid interface specified: ' + intf_name)

class Intf:
    def open(self, opts):
        pass

    def close(self):
        pass

    def issue_cmd(self, cmd_cls, *args):
        pass

    def issue_raw_cmd(self, req, lun=0):
        pass

    def issue_bridging_cmd(self, dest, target, req, lun=0):
        pass



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import time, sys, builtins
from collections import defaultdict
from . _consts import *
from .. mesg.ipmi_se import *
from .. mesg.ipmi_storage import GetSDRRepoInfo, GetSDR, RevSDR
from .. util.exception import PyExcept, PyCmdsExcept

g_add_ts = 0
g_erase_ts = 0
g_sdr_repo = defaultdict(list)
g_sensor_map = {}

conv_time = lambda ts: time.asctime(time.localtime(ts))

def bcd2str(bcd, rev=False):
    # for 1 byte only
    s1 = bcd & 0x0f
    s2 = (bcd & 0xf0) >> 4
    
    if not rev:
        ret = '{0}{1}'.format(s2, s1)        
    else: 
        ret = '{0}.{1}'.format(s1, s2)        

    return ret

def str2int(org, default=-1):
    if type(org) is not str: return default
    org = org.strip()

    try:
        if org[:2] in ('0x', '0X'):
            ret = int(org, 16)
        else:
            ret = int(org)
    except:
        ret = default
    finally:
        return ret

def get_chnl(argv):
    chnl = -1    
    if argv: chnl = str2int(argv[0])
    if chnl < 0:  chnl = 0x0e    

    return chnl

def _cal_twos_complement(val, bits):
    ret = val
    if val & (1 << (bits - 1)):  # minus        
        ret = val - (1 << bits)

    return ret

def _cal_threshold_sensor_reading(x, sdr):
    # y = L[(Mx + (B * 10^K1)) * 10^K2] units
    M = _cal_twos_complement(((sdr[20] & 0xc0) << 2) + sdr[19], 10)
    B = _cal_twos_complement(((sdr[22] & 0xc0) << 2) + sdr[21], 10)
    K1 = _cal_twos_complement(sdr[24] & 0x0f, 4)
    K2 = _cal_twos_complement(sdr[24] >> 4, 4)

    y = (M * x + (B * (10 ** K1))) * (10 ** K2)

    return y

def _cal_discrete_sensor_reading(t1):
    reading = 0
    if len(t1) >= 4:
        reading = (t1[3] << 8) + t1[2]

    return reading

def _conv_units(sdr):
    unit = sdr[15]
    base = sdr[16]
    mod = sdr[17]
    ret = ''

    if unit & 1:
        ret = 'percent'
    else:
        if base > 0 and base < len(BASE_UNITS):
            ret = BASE_UNITS[base]

        if mod > 0 and mod < len(BASE_UNITS):
            mod_str = BASE_UNITS[mod]
            unit_mod = (unit >> 1) & 3
            if unit_mod == 2:
                ret += ' / ' + mod_str
            elif unit_mod == 3:
                ret += ' * ' + mod_str

    unit_rate = (unit & 0x3f) >> 3
    if unit_rate > 0 and unit_rate < len(RATE_UNITS):
        ret += ' ' + RATE_UNITS[unit_rate]

    return ret

def _conv_sensor_name(sdr_type, sdr):
    start = (0, 43, 27, 12)[sdr_type]
    return sdr[start:].decode('latin_1')

def _conv_sensor_reading(t1, sdr_type, sdr1, with_units=True, with_raw_reading=False):
    event_reading_type = sdr1[8]
    reading_str = '0'

    if sdr_type == 1:
        if event_reading_type == 1:  # threshold-based
            reading = _cal_threshold_sensor_reading(t1[0], sdr1)                    
            if type(reading) is float:
                reading_str = '{0:.03f}'.format(reading)
            else:
                reading_str = str(reading)

            if with_units: reading_str += ' ' + _conv_units(sdr1)
        else:   # discrete
            reading = _cal_discrete_sensor_reading(t1)
            reading_str = '0x{0:04x}'.format(reading)
    else:   # sdr_type == 2
        if event_reading_type == 1:  # threshold-based
            reading_str = str(t1[0])  # couldn't calculate y = Mx + B  
            if with_units: reading_str += ' ' + _conv_units(sdr1)
        else:   # discrete
            reading = _cal_discrete_sensor_reading(t1)
            reading_str = '0x{0:04x}'.format(reading)

    if with_raw_reading:
        return (reading, reading_str)
    else:
        return reading_str

def conv_entity_id(eid):
    if eid >= 0x90 and eid <= 0xaf:
        ret = 'Chassis-Specific'
    elif eid >= 0xb0 and eid <= 0xcf:
        ret = 'Board-Set Specific'
    elif eid >= 0xd0 and eid <= 0xff:
        ret = 'OEM'
    else:
        ret = ENTITY_ID_CODES.get(eid, 'Reserved')
    
    return ret

def _conv_entity(sdr):
    entity = sdr[3]
    entity_inst = sdr[4] & 0x7f
    entity_str = '{0}.{1:02}'.format(entity, entity_inst)

    entity_name = conv_entity_id(entity)

    return (entity_str, entity_name)

def _conv_threshold(sdr, reading, opt=0):
    # opt = 0: reading
    # opt = 1: event mask
    ret = []
    events = ('lnc', 'lcr', 'lnr', 'unc', 'ucr', 'unr')
    
    if opt == 0:
        mask = (((sdr[12] >> 4) & 7) << 3) + ((sdr[10] >> 4) & 7)
    else:
        mask = 0xff

    for i in range(5):
        if mask & 1 and reading & 1:
            ret.append(events[i])
        mask >>= 1
        reading >>= 1

    return ret

def _conv_discrete(sdr, val, opt=0, output=list):
    # opt = 0: reading
    # opt = 1: event mask
    ret = []
    sensor_type = sdr[7]
    event_reading_type = sdr[8]    

    if opt == 0:
        mask = (sdr[14] << 8) + sdr[13]
    else:
        mask = 0xffff

    if event_reading_type >= 2 and event_reading_type <= 0x0c:
        # generic
        events = GENERIC_EVENT_TYPES[event_reading_type]
    elif event_reading_type == 0x6f:
        # sensor-specific
        if sensor_type in SPECIFIC_EVENT_TYPES.keys():
            events = SPECIFIC_EVENT_TYPES[sensor_type]
        else:
            return '' if output is str else []
    else:
        return '' if output is str else []

    for i in range(14):
        if mask & 1 and val & 1:
            ret.append(events[i])
        mask >>= 1
        val >>= 1

    if output is str:
        ret_str = ''
        for i in ret:
            if ret_str:  ret_str += ', ' + i
            else:  ret_str = i
                
        return ret_str
    else:
        return ret

def _conv_event_mask(sdr, opt=0):
    # opt = 0: assertion event mask
    # opt = 1: deassertion event mask
    event_reading_type = sdr[8]
    ret = []

    if event_reading_type == 1:     # threshold-based
        events = ('lnc-', 'lnc+', 'lcr-', 'lcr+', 'lnr-', 'lnr+',
                  'unc-', 'unc+', 'ucr-', 'ucr+', 'unr-', 'unr+')
        
        if opt == 0:    # assertion
            mask = ((sdr[10] & 0x3f) << 8) + sdr[9]
        else:   # deassertion
            mask = ((sdr[12] & 0x3f) << 8) + sdr[11]

        for i in range(11):
            if mask & 1: ret.append(events[i])
            mask >>= 1

    else:   # discrete
        if opt == 0:    # assertion
            mask = ((sdr[10] & 0x7f) << 8) + sdr[9]
        else:   # deassertion
            mask = ((sdr[12] & 0x7f) << 8) + sdr[11]

        ret = _conv_discrete(sdr, mask, 1)

    return ret

def conv_sensor_type(sensor_type):
    if sensor_type > 0 and sensor_type < len(SENSOR_TYPES):
        ret = SENSOR_TYPES[sensor_type]
    elif sensor_type >= 0xc0 and sensor_type <= 0xff:
        ret = 'OEM'
    else:
        ret = 'Reserved'

    return ret

def conv_event_reading(event_reading, ext=False):
    if event_reading == 1:
        ret = 'threshold' if not ext else 'Threshold'
    elif event_reading >= 2 or event_reading <= 0xc:
        ret = 'discrete' if not ext else 'Generic Discrete'
    elif event_reading == 0x6f:
        ret = 'discrete' if not ext else 'Sensor-Specific Discrete'
    elif event_reading >= 0x70 or event_reading <= 0x7f:
        ret = 'OEM'
    else:
        ret = 'n/a'

    return ret

def _conv_sensor_type_and_er(sdr):
    sensor_type = sdr[7]
    event_reading = sdr[8]

    ret = '{0:02X}h '.format(sensor_type)
    ret += conv_sensor_type(sensor_type) + ' ('
    ret +=  conv_event_reading(event_reading) + ')'

    return ret

def _conv_threshold_values(sdr1, opt=0, argv=None):
    # opt = 0: get everything from SDR
    # opt = 1: convert thresholds from input
    # opt = 2: convert hystersis from input
    thres = []

    if opt in (0, 1):
        # lnc, lc, lnr, unc, uc, unr        
        thres_sdr = (36, 35, 34, 33, 32, 31)
        m = sdr1[13] & 0x3f

        for i in range(0, 6):
            s = thres_sdr[i]
            raw = sdr1[s] if opt == 0 else argv[i]
            if m & 1: thres.append(_conv_sensor_reading((raw,), 1, sdr1, False))
            else:  thres.append('na')
            m >>= 1

    if opt == 0:
        # n_reading, n_max, n_min, pos_hys, neg_hys
        thres_sdr2 = (26, 27, 28, 37, 38)
        m2 = (1, 1, 1, sdr1[37], sdr1[38])

        for i in range(0, 5):
            s = thres_sdr2[i]
            if m2[i]: thres.append(_conv_sensor_reading((sdr1[s],), 1, sdr1, False))
            else:  thres.append('na')

    if opt == 2:
        for x in argv:
            if x: thres.append(_conv_sensor_reading((x,), 1, sdr1, False))
            else:  thres.append('na')

    return thres

def get_one_sdr(self, resv_id, next_id):
    # Get SDR: Resv ID (2) | Rec ID (2) | offset (1) | bytes_to_read (1)
    # First, read the first 5 bytes, sensor record header
    offset = 0
    b2read = 5
    sdr1 = self.intf.issue_cmd(GetSDR, resv_id, next_id, offset, b2read)

    # Then, read the following bytes.  Each time read 32 bytes at most
    offset = 5
    b2read = sdr1[6]
    while b2read > 0:
        if b2read > 32:
            to_read = 32
            b2read -= 32
        else:
            to_read = b2read
            b2read = 0
        rsp1 = self.intf.issue_cmd(GetSDR, resv_id, next_id, offset, to_read)
        sdr1 += rsp1[2:]
        offset += to_read

    return sdr1

def load_sdr_repo(self):
    next_id = b'\0'
    sdr_count = 0

    def load_one_sdr():
        nonlocal next_id, sdr_count

        resv_id, = self.intf.issue_cmd(RevSDR)
        sdr1 = get_one_sdr(self, resv_id, next_id)
        next_id = sdr1[:2]
        rec_type = sdr1[5]
        g_sdr_repo[rec_type].append(sdr1[7:])
        sdr_count += 1

        # Handle sensor map
        if rec_type in (1, 2, 3):
            sdr2 = sdr1[7:]
            sensor_num = sdr2[2]    # Sensor number
            sensor_name = _conv_sensor_name(rec_type, sdr2)  # Sensor name

            entity_str, entity_name, units, sensor_type = '', '', '', ''
            thres = ['na'] * 11
            mask_r, mask_s, asserts, deasserts = [], [], [], []
            mask = sdr2[13] & 0x3f

            if rec_type in (1, 2):
                event_reading_type = sdr2[8]
                entity_str, entity_name = _conv_entity(sdr2)
                units = _conv_units(sdr2)
                sensor_type = _conv_sensor_type_and_er(sdr2)
                asserts = _conv_event_mask(sdr2, 0)
                deasserts = _conv_event_mask(sdr2, 1)

                if event_reading_type == 1:
                    mask_r = _conv_threshold(sdr2, mask, 1)
                    mask_s = _conv_threshold(sdr2, sdr2[14] & 0x3f, 1)

                    if rec_type == 1:
                        thres = _conv_threshold_values(sdr2)
                else:
                    mask_r = _conv_discrete(sdr2, ((sdr2[14] & 0x7f) << 8) + sdr2[13], 1)

            val = (sensor_name, rec_type, len(g_sdr_repo[rec_type]) - 1, 
                    entity_str, entity_name, units, thres, mask_r, mask_s,
                    sensor_type, asserts, deasserts)

            if not g_sensor_map.get(sensor_num, []):
                g_sensor_map[sensor_num] = [val]
            else:
                g_sensor_map[sensor_num].append(val)

    global g_add_ts, g_erase_ts, g_sensor_map
    _, rec, _, add_ts, erase_ts, _ = self.intf.issue_cmd(GetSDRRepoInfo)
    if rec == 0:  
        raise PyCmdsExcept('SDR repository is empty.', -1)

    if g_sdr_repo:
        if add_ts == g_add_ts and erase_ts == g_erase_ts:
            # no change on the SDR repository
            return False
        else:
            # SDR has changes and it's not empty
            # clear the previous SDRs before getting the new ones
            g_sdr_repo.clear()
            g_sensor_map.clear()

    #self.print('Loading {0} SDRs...\n'.format(rec))

    load_one_sdr()
    while next_id != b'\xff\xff':
        if sdr_count > rec: break   # already got more SDRs than the total # 
        load_one_sdr()

    g_add_ts = add_ts
    g_erase_ts = erase_ts
    g_sensor_map = dict(sorted(g_sensor_map.items()))

    return True

def get_sdr_repo(self):
    load_sdr_repo(self)
    return g_sdr_repo

def get_sensor_map(self):
    load_sdr_repo(self)
    return g_sensor_map    

def get_sensor_readings(self, opt=1, filter_sdr=0, filter_sensor_type=0, ext=False,
                        filter_sensor_num=-1):
    load_sdr_repo(self)
    keys = g_sensor_map.keys()
    if filter_sensor_num != -1:
        if not filter_sensor_num in keys:
            raise PyCmdsExcept('Sensor number: {0:02X}h was not found.'.format(
                               filter_sensor_num), -1)
        keys = (filter_sensor_num,)

    for sensor_num in keys:
        for rec in g_sensor_map[sensor_num]:
            (sensor_name, sdr_type, idx, entity_str, entity_name, units, 
             thres, mask_r, mask_s, sensor_type, asserts, deasserts) = rec

            if sdr_type == 3:   continue    # Do not handle Event-Only SDRs here

            sdr1 = g_sdr_repo[sdr_type][idx]
            event_reading_type = sdr1[8]

            # filter SDR type
            if filter_sdr != 0 and filter_sdr != sdr_type:
                continue

            # filter sensor type
            if filter_sensor_type != 0 and filter_sensor_type != sdr1[7]:
                continue

            # IPMI: Get sensor reading 
            try:
                t1 = self.intf.issue_cmd(GetSensorReading, sensor_num)
            except:
                continue

            # Sensor status
            stat = (t1[1] & 0x7f) >> 5
            stat_str = ('ns', 'na', 'ok', 'ns')[stat]

            # Convert the sensor reading to human readable format
            reading = 0
            reading_str = ''

            if stat == 2:
                if opt == 3:
                    reading_str = _conv_sensor_reading(t1, sdr_type, sdr1, False)
                else:
                    reading, reading_str = _conv_sensor_reading(t1, sdr_type, sdr1, True, True)                    
            elif stat & 2 == 0:
                reading_str = 'disabled'
            else:
                reading_str = 'not available'

            if opt in (1, 2):    # sdr list, sdr elist
                if opt == 2:
                    if stat == 2 and event_reading_type != 1:
                        reading_str = _conv_discrete(sdr1, reading, 0, str)

                yield (sensor_num, sensor_name, reading_str, stat_str, entity_str)
                continue

            # The remaining is for opt == 3 or 4, i.e. sdr slist/vlist or sensor list/vlist        
            if ext:  # In ext mode, overwrite the thresholds from SDR with the ones from command
                try:
                    # Get Sensor Thresholds command
                    argv = self.intf.issue_cmd(GetSensorThres, sensor_num)
                    thres[:6] = _conv_threshold_values(sdr1, 1, argv[1:])
                except:
                    pass

            if opt == 3:  # sdr slist or sensor list
                ret = [sensor_num, sensor_name, reading_str, units, stat_str,] + thres[:6]
                yield ret
                continue

            # The remaining is all for opt == 4, i.e. sdr vlist or sensor vlist
            if ext:  # In ext mode, overwrite the hysteresis from SDR with the ones from command
                try:
                    # Get Sensor Hysteresis command
                    argv = self.intf.issue_cmd(GetSensorHys, sensor_num)
                    thres[9:] = _conv_threshold_values(sdr1, 2, argv)
                except:
                    pass

            # Get absolute values of Hysteresis
            thres[9:] = [x[1:] if x[0] == '-' else x for x in thres[9:]]

            asserted_events = []
            if stat == 2:
                if event_reading_type != 1:
                    asserted_events = _conv_discrete(sdr1, reading)
                else:
                    if len(t1) >= 3:
                        asserted_events = _conv_threshold(sdr1, t1[2] & 0x3f) 
                
            ret = [sensor_num, sensor_name, entity_str, entity_name, sensor_type, reading_str, stat_str,]
            ret += thres
            ret += [asserts, deasserts, event_reading_type, asserted_events, mask_r, mask_s,]
                            
            yield ret

def print_sensor_list1(self, reading_all):
    for sensor_num, sensor_name, reading, stat, _ in reading_all:
        fmt = '{0:02X}h | {1:16} | {2:16} | {3}' 
        self.print(fmt.format(sensor_num, sensor_name, reading, stat))

def print_sensor_list2(self, reading_all):
    for num, name, reading, stat, entity in reading_all:
        fmt = '{0:02X}h | {1:16} | {3:>5} | {2} | {4}' 
        self.print(fmt.format(num, name, stat, entity, reading))

def print_sensor_list3(self, reading_all):
    for (sensor_num, sensor_name, reading, unit, stat, 
         lnc, lc, lnr, unc, uc, unr) in reading_all:
        if reading == 'not available':  reading = 'na'
        if reading == 'disabled':  reading = 'ns'

        fmt = '{0:02X}h | {1:16} | {2:8} | {3:10} | {4}' 
        fmt += ' | {7:7} | {6:7} | {5:7} | {8:7} | {9:7} | {10:7}'

        self.print(fmt.format(sensor_num, sensor_name, reading, unit, stat,
                         lnc, lc, lnr, unc, uc, unr))

def print_sensor_list4(self, reading_all):
    def print_threshold_events(title, events):
        if events:
            self.print(title, end=' ')
            for i in events:
                self.print(i, end=' ')
            self.print('')

    def print_discrete_events(title, events):
        if events:
            self.print(title + ' [' + events[0] + ']')
            for i in events[1:]:
                self.print(' ' * 25 + '[' + i + ']')

    for (sensor_num, sensor_name, entity_str, entity_name, sensor_type, 
         reading_str, stat_str, lnc, lc, lnr, unc, uc, unr, n_reading, n_max, n_min,
         pos_hys, neg_hys, asserts, deasserts, er_type, asserted_events, 
         mask_r, mask_s) in reading_all:

        self.print('Sensor ID              : {0} ({1:02X}h)'.format(sensor_name, sensor_num))
        self.print('  Entity ID            : ' + entity_str + ' (' + entity_name + ')')        
        self.print('  Sensor Type          : ' + sensor_type)
        self.print('  Sensor Reading       : ' + reading_str)
        self.print('  Status               : ' + stat_str)

        if er_type == 1:    # threshold
            self.print('  Nominal Reading      : ' + n_reading)
            self.print('  Normal Minimum       : ' + n_min)
            self.print('  Normal Maximum       : ' + n_max)
            if unr != 'na': self.print('  Upper Non-Recoverable: ' + unr)
            if uc != 'na':  self.print('  Upper Critical       : ' + uc)
            if unc != 'na': self.print('  Upper Non-Critical   : ' + unc)
            if lnc != 'na': self.print('  Lower Non-Critical   : ' + lnc)
            if lc != 'na':  self.print('  Lower Critical       : ' + lc)
            if lnr != 'na': self.print('  Lower Non-recoverable: ' + lnr)
            if pos_hys != 'na': self.print('  Positive Hysteresis  : ' + pos_hys)
            if neg_hys != 'na': self.print('  Negative Hysteresis  : ' + neg_hys)
            
            print_func = print_threshold_events
            print_func('  Readable Thresholds  :', mask_r)
            print_func('  Settable Thresholds  :', mask_s)

        else:   # discrete
            print_func = print_discrete_events
            print_func('  Reading Mask         :', mask_r)

        print_func('  Asserted Events      :', asserted_events)
        print_func('  Assertions Enabled   :', asserts)
        print_func('  Deassertions Enabled :', deasserts)

        self.print(' ')

def print_sensor_type_list(self):
    self.print('Sensor Types:')
    self.print('    Temperature               (0x01)   Voltage                   (0x02)')
    self.print('    Current                   (0x03)   Fan                       (0x04)')
    self.print('    Physical Security         (0x05)   Platform Security         (0x06)')
    self.print('    Processor                 (0x07)   Power Supply              (0x08)')
    self.print('    Power Unit                (0x09)   Cooling Device            (0x0a)')
    self.print('    Other                     (0x0b)   Memory                    (0x0c)')
    self.print('    Drive Slot / Bay          (0x0d)   POST Memory Resize        (0x0e)')
    self.print('    System Firmwares          (0x0f)   Event Logging Disabled    (0x10)')
    self.print('    Watchdog1                 (0x11)   System Event              (0x12)')
    self.print('    Critical Interrupt        (0x13)   Button                    (0x14)')
    self.print('    Module / Board            (0x15)   Microcontroller           (0x16)')
    self.print('    Add-in Card               (0x17)   Chassis                   (0x18)')
    self.print('    Chip Set                  (0x19)   Other FRU                 (0x1a)')
    self.print('    Cable / Interconnect      (0x1b)   Terminator                (0x1c)')
    self.print('    System Boot Initiated     (0x1d)   Boot Error                (0x1e)')
    self.print('    OS Boot                   (0x1f)   OS Critical Stop          (0x20)')
    self.print('    Slot / Connector          (0x21)   System ACPI Power State   (0x22)')
    self.print('    Watchdog2                 (0x23)   Platform Alert            (0x24)')
    self.print('    Entity Presence           (0x25)   Monitor ASIC              (0x26)')
    self.print('    LAN                       (0x27)   Management Subsys Health  (0x28)')
    self.print('    Battery                   (0x29)   Session Audit             (0x2a)')
    self.print('    Version Change            (0x2b)   FRU State                 (0x2c)')

def do_command(self, argv, help_func, cmd_map):
    if type(argv) is str:
        argv = argv.split(' ')
        argv = [x.strip() for x in argv if x.strip() != '']

    try:
        if len(argv) == 0:
            raise PyCmdsExcept('No input command.')

        if argv[0] not in cmd_map.keys():
            raise PyCmdsExcept('Command ' + argv[0] + ' is not supported.')

        # Pass the arguments to the corresponding command handler
        cmd_map[argv[0]](self, argv)    

    except PyCmdsExcept as e:
        builtins.print(e, '\n')
        if e.context >= 0:  help_func(self, argv, e.context)

    except BaseException as e:
        builtins.print(e)

def print(self, *objects, sep=' ', end='\n', flush=False):
    builtins.print(*objects, sep=sep, end=end, file=self.print_file, flush=flush)

def print_rsp(self, rsp):
    self.print(' '.join(('{0:02x}'.format(i) for i in rsp)))



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import struct
from . _crypto import *
from .. util import checksum
from .. util.exception import PyIntfExcept, PyIntfSeqExcept

class RMCP_Message:
    SOFT_ID = 0x81
    BMC_ADDR = 0x20

    def __init__(self, msg_cls=6, seq_num=0xff):
        self.msg_cls = msg_cls
        self.rmcp_seq_num = seq_num

    def pack(self):
        msg = struct.pack('BBBB', 6, 0, self.rmcp_seq_num, self.msg_cls)
        return msg

    def unpack(self, rsp):
        if rsp[:4] != struct.pack('BBBB', 6, 0, 0xff, self.msg_cls):
            raise PyIntfExcept('Invalid RMCP package.')

        return rsp[4:]

class ASF_Ping(RMCP_Message):
    seq_num = 1
    def __init__(self):
        # Generate the RMCP header
        self.seq_num = ASF_Ping.seq_num        
        super(ASF_Ping, self).__init__(seq_num=self.seq_num)
        
        seq_num = self.seq_num
        seq_num += 1
        if seq_num == 0xff: seq_num = 0
        RMCP_Message.seq_num = seq_num

    def pack(self):
        msg = super(ASF_Ping, self).pack()
        msg_ping = struct.pack('>LBBBB', 4542, 0x80, self.seq_num, 0, 0)
        msg += msg_ping

        return msg

    def unpack(self, rsp):
        pass

class IPMI15_Message(RMCP_Message):
    seq_num = 1

    def __init__(self, auth=AUTH_NONE, sseq=0, sid=b'\0', pwd=''):
        # Get this seq_num
        if isinstance(self, IPMI15_Message):
            self.seq_num = IPMI15_Message.seq_num
            IPMI15_Message.seq_num += 1
            if IPMI15_Message.seq_num == 64: IPMI15_Message.seq_num = 0

        self.auth = auth
        self.sseq = sseq
        self.sid = sid
        self.pwd = pwd
        self.rs_addr = RMCP_Message.BMC_ADDR
        
        # Generate the RMCP header
        super(IPMI15_Message, self).__init__(msg_cls=7)

    def _pack_lan_payload(self, req, rs_addr=RMCP_Message.BMC_ADDR):
        # rsAddr (SA or SW ID) | NetFn (even) / rsLUN | checksum
        p1 = struct.pack('<BB', rs_addr, (req.netfn << 2) + req.lun)                     
        p1 += checksum(p1)

        # rqAddr (SA or SW ID)  | rqSeq / rqLUN | cmd | req data (0..N) | checksum
        p2 = struct.pack('<BBB', RMCP_Message.SOFT_ID, self.seq_num << 2, req.cmd)
        if req.req_data != None: p2 += req.req_data
        p2 += checksum(p2)

        return p1 + p2

    def _pack_lan_header(self, payload):
        payload_len = len(payload)
        if self.auth == AUTH_NONE:
            # Auth Type (1) | Seq # (4) | Sess ID (4) | Payload Len (1)
            hdr = struct.pack('<BL4sB', self.auth, self.sseq, self.sid, payload_len)
        else:
            # Auth Type (1) | Seq # (4) | Sess ID (4) |  AuthCode (16) | Payload Len (1)
            auth_code = cal_auth_code_15(self.auth, self.pwd, self.sseq, self.sid, payload)

            hdr = struct.pack('<BL4s16sB', self.auth, self.sseq, self.sid, auth_code, 
                              payload_len)   

        return hdr

    def pack(self, req):
        # Pack IPMI Payload
        payload = self._pack_lan_payload(req)
        if len(payload) > 255:
            raise PyIntfExcept('Request data exceeds max length.')

        # pack the IPMI LAN 1.5 header
        hdr = self._pack_lan_header(payload)
        
        # pack RMCP header
        msg = super(IPMI15_Message, self).pack()
        msg += hdr + payload

        return msg
    
    def _unpack_lan_payload(self, payload):
        # verify checksum        
        if (checksum(payload[:2], int) != payload[2] or
            checksum(payload[3:len(payload)-1], int) != payload[-1]):
            raise PyIntfExcept('Invalid RMCP checksum in response.')        

        # unpack data fields
        rq_addr, netfn = struct.unpack('BB', payload[:2])
        netfn >>= 2
        rs_addr, rq_seq, cmd, cc = struct.unpack('BBBB', payload[3:7])
        rq_seq >>= 2
        rsp_data = payload[7:len(payload)-1]

        if rq_addr != RMCP_Message.SOFT_ID:
            raise PyIntfExcept('Invalid rq_addr: {0:02X}h in response.  Expected {1:02X}h.'.format(
                               rq_addr, RMCP_Message.SOFT_ID)) 

        if rs_addr != self.rs_addr:
            raise PyIntfExcept('Invalid rs_addr: {0:02X}h in response.  Expected {1:02X}h.'.format(
                               rs_addr, self.rs_addr)) 

        if rq_seq != self.seq_num:
            raise PyIntfSeqExcept('RMCP sequence number mismatches in response.')  

        return (netfn, cmd, cc, rsp_data)

    def _unpack_lan_header(self, rsp):
        # unpack payload
        try:
            if self.auth == AUTH_NONE:
                # Auth Type (1) | Seq # (4) | Sess ID (4) | Payload Len (1)
                t1 = struct.unpack('<BL4sB', rsp[:10])
                payload = rsp[10:]
            else:
                # Auth Type (1) | Seq # (4) | Sess ID (4) |  AuthCode (16) | Payload Len (1)
                t1 = struct.unpack('<BL4s16sB', rsp[:26])
                payload = rsp[26:]
        except:
            raise PyIntfExcept('Invalid RMCP header in response.')        

        # unpack IPMI LAN package
        if len(payload) != t1[-1] or len(payload) < 7:
            raise PyIntfExcept('Invalid RMCP payload length in response.')        

        if t1[0] != self.auth:
            raise PyIntfExcept('Invalid RMCP header in response.')        

        # verify AuthCode
        if t1[0] != AUTH_NONE:
            auth_code = cal_auth_code_15(t1[0], self.pwd, t1[1], t1[2], payload)
            if auth_code != t1[3]:
                raise PyIntfExcept('Invalid RMCP AuthCode in response.')  
        
        return payload

    def unpack(self, rsp):
        # unpack RMCP header
        rsp = super(IPMI15_Message, self).unpack(rsp)

        # unpack IPMI 1.5 header
        payload = self._unpack_lan_header(rsp)
                    
        # unpack IPMI LAN package
        return self._unpack_lan_payload(payload)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, struct, socket, select, threading, time
from .. util.exception import PyIntfExcept, PyIntfSeqExcept

from . import Intf

from . _rmcp_msg import ASF_Ping, IPMI15_Message
from . _crypto import RMCP_AUTHS, AUTH_NONE, conv_str2bytes

from .. mesg import IPMI_Raw
from .. mesg.ipmi_app import IPMI_SendMsg, GetChnlAuthCap,\
        GetSessChallenge, ActivateSess, SetSessPriv, CloseSess

class RMCP_Ping(Intf):
    def __init__(self, opts):
        self.socket = None
        self.host = opts.get('host', 'localhost')
        self.port = opts.get('port', 623)

    def __del__(self):
        self.close()

    def open(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            self.socket.connect((self.host, self.port))
        except:
            raise PyIntfExcept("Failed to connect to host.")

    def close(self):
        try:
            if self.socket:
                self.socket.close()
                self.socket = None
        except:
            pass

    def sendrecv(self, data, retries=3):
        while retries:
            self.socket.send(data)
            r, _, x = select.select([self.socket], [], [], 1.5)
            if x:  raise PyIntfExcept('Socket exception occurred.  Stopped.')
            if r:  return self.socket.recv(4096)
            retries -= 1

        raise PyIntfExcept('Times out.  Host has no response.')

    def ping(self):
        ping = ASF_Ping()
        msg = ping.pack()
        rsp = self.sendrecv(msg)
        if len(rsp) != 28:
            raise PyIntfExcept('Host responded improperly.')

class RMCP(RMCP_Ping):
    KEEP_ALIVE_PERIOD = 55

    def __init__(self, opts, keep_alive):
        self.auth = AUTH_NONE
        self.sseq = 0
        self.sid = b'\0'
        self.sess_act = False

        self.keep_alive = keep_alive
        self.wd_count = RMCP.KEEP_ALIVE_PERIOD
        self.lock = threading.Lock()
        
        super(RMCP, self).__init__(opts)

    def __del__(self):
        self.close()

    def open(self, opts):
        if 'auth' in opts.keys() and opts['auth'] not in RMCP_AUTHS.keys():
            raise PyIntfExcept('Authentication algorithm {0} is not supported.'
                                 .format(opts['auth']))

        super(RMCP, self).open()

        user = conv_str2bytes(opts.get('user', None))
        self.passwd = conv_str2bytes(opts.get('password', None))
        priv = opts.get('priv', 4)        
        self.ioseq = 0

        no_ping = opts.get('no_ping', False)
        if not no_ping:        
            # rmcp ping
            self.ping()

        # open session
        # Get Channel Authentication Capabilities Command
        self.issue_cmd(GetChnlAuthCap, priv)

        # Get Session Challenge Command
        auth = RMCP_AUTHS[opts.get('auth', 'md5')]
        self.sid, chg_data = self.issue_cmd(GetSessChallenge, auth, user)

        # Activate Session Command
        self.auth = auth
        while self.ioseq == 0:
            self.ioseq, = struct.unpack('<L', os.urandom(4))
        self.auth, self.sid, self.sseq, priv = self.issue_cmd(
                                     ActivateSess, auth, priv, chg_data, self.ioseq)

        self.sess_act = True  # Mark the session as activated

        # Set Session Privilege Level Command
        priv, = self.issue_cmd(SetSessPriv, priv)

        # Create the keep-alive thread
        if self.keep_alive:
            th = threading.Thread(target=self.keep_alive_cb)
            th.daemon = True
            th.start()

    def close(self):
        try:
            if self.sess_act:
                # Close Session Command
                self.issue_cmd(CloseSess, self.sid) # close session
                self.sess_act = False
        except:
            pass
        finally:
            super(RMCP, self).close() # close socket

    def recv(self):
        r, _, x = select.select([self.socket], [], [], 1.5)
        if x:   raise PyIntfExcept('Socket exception occurred.  Stopped.')
        if r:   return self.socket.recv(4096)
        return None

    def gen_msg(self, cmd, bridging=False, dest=0, target=0):
        msg = IPMI15_Message(self.auth, self.sseq, self.sid, self.passwd)

        if bridging:
            # Message bridging
            inner = msg._pack_lan_payload(cmd, target)
            sm = IPMI_SendMsg(dest, inner)
            data = msg.pack(sm)
            ret = (msg, data, sm)
        else:
            # Common IPMI commands
            data = msg.pack(cmd)
            ret = (msg, data)

        return ret

    def unpack(self, rsp, msg, cmd):
        pkt1 = msg.unpack(rsp)  
        pkt2 = cmd.unpack(pkt1)
        return pkt2

    def _issue_cmd_imp(self, cmd_cls, *args):
        if self.keep_alive: self.wd_count = RMCP.KEEP_ALIVE_PERIOD
        cmd = cmd_cls(*args)

        if self.sess_act:
            self.sseq += 1
            if self.sseq > 0xffffffff:
                self.sseq = 1

        msg, data = self.gen_msg(cmd)
        rsp = self.sendrecv(data)

        try:
            return self.unpack(rsp, msg, cmd)   

        except PyIntfSeqExcept:
            # The response mismatches the request
            # Check the next response message
            retries = 2
            while retries:
                try:
                    rsp = self.recv()
                    if rsp: return self.unpack(rsp, msg, cmd)

                except PyIntfSeqExcept:
                    pass

                finally:
                    retries -= 1

        raise PyIntfExcept('Could not match the request with the response message.')        

    def _issue_bridging_cmd_imp(self, dest, target, req, lun):
        if self.keep_alive: self.wd_count = RMCP.KEEP_ALIVE_PERIOD
        cmd = IPMI_Raw(req, lun)

        if self.sess_act:
            self.sseq += 1
            if self.sseq > 0xffffffff:
                self.sseq = 1

        msg, data, cmd_sm = self.gen_msg(cmd, True, dest, target)
        rsp = self.sendrecv(data, 1)

        # The 1st response of send message received
        try:
            self.unpack(rsp, msg, cmd_sm)   
        except PyIntfSeqExcept:
            # The version of HS9216 AMI I tested has a bug of returning the wrong 
            # seq number in the 1st response.  It occurs around 1%.
            # Here is for working around the bug
            pass

        retries = 3
        while retries:
            rsp = self.recv() 
            if rsp:  break
            retries -= 1
        
        if not rsp:         
            raise PyIntfExcept('Message Bridging times out.')        

        # The 2nd response of the inner bridged message received
        msg.rs_addr = target
        return self.unpack(rsp, msg, cmd)

    def issue_cmd(self, cmd_cls, *args):
        with self.lock:
            rsp = self._issue_cmd_imp(cmd_cls, *args)
        return rsp

    def issue_raw_cmd(self, req, lun=0):
        return self.issue_cmd(IPMI_Raw, req, lun)
    
    def issue_bridging_cmd(self, dest, target, req, lun=0):
        with self.lock:
            rsp = self._issue_bridging_cmd_imp(dest, target, req, lun)        
        return rsp
    
    def keep_alive_cb(self):
        while True:
            if self.wd_count == 0 and self.sess_act:
                self.issue_raw_cmd([6, 1])
            self.wd_count -= 1
            time.sleep(1)
            



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from .. mesg.ipmi_transport import *
from . _common import str2int, get_chnl, do_command
from . _consts import USER_PRIVS, AUTH_TYPES
from .. util.exception import PyCmdsExcept, PyCmdsArgsExcept

def _conv_auth_types(val):
    ret = []
    auths = list(AUTH_TYPES.keys())

    idx = 0
    while val:
        if val & 1: ret.append(auths[idx])
        idx += 1
        val >>= 1
        if idx == 3:  val >>= 1
        if idx >= len(auths):  break

    return ret

def _print_auth_types(self, fmt, title, val):    
    self.print(fmt.format(title), end=' ')
    ret = _conv_auth_types(val)
    if not ret:
        self.print('(all disabled)')
    else:
        self.print(' '.join(ret))

def _print_ip_addr(self, title, val):
    if len(val) < 4:
        raise PyCmdsExcept('Invalid IP Address returned by BMC.', -1)
    self.print('{0} {1}.{2}.{3}.{4}'.format(title, val[0], val[1], val[2], val[3]))

def _print_mac_addr(self, title, val):
    if len(val) < 6:
        raise PyCmdsExcept('Invalid MAC Address returned by BMC.', -1)
    self.print('{0} {1:02x}:{2:02x}:{3:02x}:{4:02x}:{5:02x}:{6:02x}'.format(
         title, val[0], val[1], val[2], val[3], val[4], val[5]))

def _print_ipv6_addr(self, title, val):
    if len(val) < 16:
        raise PyCmdsExcept('Invalid IPv6 Address returned by BMC.', -1)
    self.print(title, end=' ')
    for i in range(7):
        self.print('{0:02x}{1:02x}:'.format(val[i*2], val[i*2+1]), end='')
    self.print('{0:02x}{1:02x}'.format(val[14], val[15]))    

def _lan_print_1(self, rsp):
    # 1 Auth Type Support
    rsp = int.from_bytes(rsp, 'little') 
    _print_auth_types(self, '{0:22}:', 'Auth Type Support', rsp)

def _lan_print_2(self, rsp):
    # 2 Authentication Type Enables
    rsp = list(rsp)
    self.print('Auth Type Enables')
    for i in range(5):
        _print_auth_types(self, '  {0:20}:', USER_PRIVS[i + 1] , rsp[i])

def _lan_print_3(self, rsp):
    # 3 IP Address
    rsp = list(rsp)
    _print_ip_addr(self, 'IP Address            :', rsp)

def _lan_print_4(self, rsp):
    # 4 IP Address Source
    rsp = int.from_bytes(rsp, 'little') 
    src = {0: 'unspecified', 1: 'static address', 2: 'dhcp', 3: 'loaded by host',
           4: 'others'}.get(rsp, '(unknown)')
    self.print('IP Address Source     :', src)

def _lan_print_5(self, rsp):
    # 5 MAC Address
    rsp = list(rsp)
    _print_mac_addr(self, 'MAC Address           :', rsp)

def _lan_print_6(self, rsp):
    # 6 Subnet Mask
    rsp = list(rsp)
    _print_ip_addr(self, 'Subnet Mask           :', rsp)

def _lan_print_7(self, rsp):
    # 7 IPv4 Header Parameters
    rsp = list(rsp)
    ttl = rsp[0]
    flags = (rsp[1] & 0xe0) >> 5
    precedence = (rsp[2] & 0xe0) >> 5
    tos = (rsp[2] & 0x1e) >> 1
    self.print('IPv4 Header           : TTL=0x{0:02x} Flags=0x{1:02x} Precedence=0x{2:02x} TOS=0x{3:02x}'.format(
          ttl, flags, precedence, tos))

def _lan_print_10(self, rsp):
    # 10 BMC-generated ARP control
    rsp = int.from_bytes(rsp, 'little') 
    arp1 = 'Enabled' if rsp & 2 else 'Disabled'
    arp2 = 'Enabled' if rsp & 1 else 'Disabled'

    self.print('BMC ARP Control       : ARP Responses {0}, Gratuitous ARP {1}'.format(arp1, arp2))

def _lan_print_12(self, rsp):
    # 12 Default Gateway Address
    rsp = list(rsp)
    _print_ip_addr(self, 'Default Gateway IP    :', rsp)

def _lan_print_13(self, rsp):
    # 13 Default Gateway MAC Address
    rsp = list(rsp)
    _print_mac_addr(self, 'Default Gateway MAC   :', rsp)

def _lan_print_14(self, rsp):
    # 14 Default Gateway Address
    rsp = list(rsp)
    _print_ip_addr(self, 'Backup Gateway IP     :', rsp)

def _lan_print_15(self, rsp):
    # 15 Backup Gateway MAC Address
    rsp = list(rsp)
    _print_mac_addr(self, 'Backup Gateway MAC    :', rsp)

def _lan_print_20(self, rsp):
    # 20 802.1q VLAN ID
    rsp = list(rsp)
    if rsp[1] & 0x80 == 0:
        ret = 'disabled'
    else:
        ret = ((rsp[1] & 0x0f) << 8) + rsp[0]
        ret = '{0:03x}'.format(ret)

    self.print('802.1q VLAN ID        : {0}'.format(ret))

def _lan_print_21(self, rsp):
    # 21 802.1q VLAN Priority
    rsp = int.from_bytes(rsp, 'little') 
    self.print('802.1q VLAN Priority  : {0}'.format(rsp & 7))

def _lan_get_ciphers(self, chnl):
    # 22 RMCP+ Messaging Cipher Suite Entry Support
    rsp = self.intf.issue_cmd(GetLanConfig, chnl, 22, 0, 0)
    if not rsp: 
        raise PyCmdsExcept('Failed to get the number of suppported RMCP+ Cipher Suites.', -1)

    num = int.from_bytes(rsp, 'little') 
    if not num:
        raise PyCmdsExcept('The number of supported RMCP+ Cipher Suites is zero.', -1)

    # 23 RMCP+ Messaging Cipher Suite Entries
    entries = self.intf.issue_cmd(GetLanConfig, chnl, 23, 0, 0)
    if not entries:    
        raise PyCmdsExcept('Failed to get any supported RMCP+ Cipher Suites.', -1)

    entries = list(entries)

    # 24 RMCP+ Messaging Cipher Suite Privilege Levels
    privs = self.intf.issue_cmd(GetLanConfig, chnl, 24, 0, 0)
    if not privs:
        raise PyCmdsExcept('Failed to get any privilege levels of the supported RMCP+ Cipher Suites.', -1)

    privs = list(privs)

    return (num, entries, privs)

def _lan_print_ciphers(self, num, entries, privs):
    self.print('RMCP+ Cipher Suites   : ', end='')
    self.print(','.join((str(x) for x in entries[1:])))

    try:
        count = (num + 1) / 2 if num & 1 else num / 2
        count = int(count)
        self.print('Cipher Suite Priv Max : ', end='')
        for i in range(1, count):
            self.print('{0},{1},'.format(privs[i] & 0x0f, (privs[i] & 0xf0) >> 4), end='')
    
        if num & 1:
            self.print(privs[count] & 0x0f)
        else:
            self.print('{0},{1}'.format(privs[count] & 0x0f, (privs[count] & 0xf0) >> 4))
    except:
        raise PyCmdsExcept('Unexpected privilege levels of the supported RMCP+ Cipher Suites returned.', -1)

    finally:
        self.print(' ' * 23 + '* 0h = Unused')
        for i in range(1, 6):
            self.print(' ' * 23 + '* {0}h = {1}'.format(i, USER_PRIVS[i]))

LAN_PRINT_HDL = {
    1: (_lan_print_1, 1),
    2: (_lan_print_2, 5),
    3: (_lan_print_3, 4),
    4: (_lan_print_4, 1),
    5: (_lan_print_5, 6),
    6: (_lan_print_6, 4),
    7: (_lan_print_7, 3),
    10: (_lan_print_10, 1),
    12: (_lan_print_12, 4),
    13: (_lan_print_13, 6),
    14: (_lan_print_14, 4),
    15: (_lan_print_15, 6),
    20: (_lan_print_20, 2),
    21: (_lan_print_21, 1),
}

def _lan_print(self, argv):
    chnl = get_chnl(argv[1:])
    flag = False
    for key in LAN_PRINT_HDL.keys():
        rsp = self.intf.issue_cmd(GetLanConfig, chnl, key, 0, 0)
        if not rsp:  continue

        print_func = LAN_PRINT_HDL[key][0]
        rsp_len = LAN_PRINT_HDL[key][1]
        if len(rsp) != rsp_len:
            continue
        
        flag = True
        print_func(self, rsp)

    num, entries, privs = _lan_get_ciphers(self, chnl)
    flag = True
    _lan_print_ciphers(self, num, entries, privs)

    if not flag:
        raise PyCmdsExcept('Failed to get any LAN configuration parameters.', -1)

def _lan_alert(self, argv):
    chnl = get_chnl(argv[1:])

    # 17 Number of Destinations
    rsp = self.intf.issue_cmd(GetLanConfig, chnl, 17, 0, 0)
    if not rsp: 
        raise PyCmdsExcept('Failed to get the number of alert destinations.', -1)
    
    num = int.from_bytes(rsp, 'little') 
    if not num:
        raise PyCmdsExcept('The number of alert destinations is zero.', -1)

    flag = False
    for i in range(0, num + 1):
        # 18 Destination Type
        rsp = self.intf.issue_cmd(GetLanConfig, chnl, 18, i, 0)
        if not rsp: continue
        
        # 19 Destination Addresses
        rsp2 = self.intf.issue_cmd(GetLanConfig, chnl, 19, i, 0)
        if not rsp2: continue
        
        flag = True
        rsp = list(rsp)
        rsp2 = list(rsp2)

        # Output results
        ack = 'Acknowledged' if rsp[1] & 0x80 else 'Unacknowledged'
        dest_type = {0: 'PET Trap', 6: 'OEM 1', 7: 'OEM 2'}.get(rsp[1] & 7, 'Unknown')

        self.print('Alert Destination      : {0}'.format(rsp[0]))
        self.print('Alert Acknowledge      : {0}'.format(ack))
        self.print('Destination Type       : {0}'.format(dest_type))
        if rsp[1] & 0x80:
            self.print('Acknowledge Timeout    : {0} seconds'.format(rsp[2]))
        self.print('Number of Retries      : {0}'.format(rsp[3] & 7))

        if rsp2[1] & 0x10:   # IPv6
            _print_ipv6_addr(self, 'Alert IPv6 Address     :', rsp2[2:])
        else:
            gateway = 'default' if rsp2[2] & 1 == 0 else 'backup'

            self.print('Alert Gateway          : {0}'.format(gateway))
            _print_ip_addr(self, 'Alert IP Address       :', rsp2[3:7])
            _print_mac_addr(self, 'Alert MAC Address      :', rsp2[7:])
            
        self.print('')

    if not flag:
        raise PyCmdsExcept('Failed to get any alert destinations.', -1)

def _lan_enable_cipher(self, argv, op=1):
    # op = 0: disable a cipher suite
    # op = 1: enable a cipher suite

    chk = 1 if op == 0 else 2
    if len(argv) < chk:
        raise PyCmdsArgsExcept(1, 1)

    cipher = str2int(argv[0])
    if cipher < 0:
        raise PyCmdsExcept('Invalid cipher suite: ' + argv[0], 1)

    if op == 1:  # enable
        priv = str2int(argv[1])
        if priv < 1 or priv > 5:
            raise PyCmdsExcept('Invalid privilege level: ' + argv[1], 1)
    else:  # disable
        # prohibit to disable the in-use cipher suite
        if cipher == self.intf.cipher_suite:
            raise PyCmdsExcept('Error: Cipher Suite {0} is in use.'.format(cipher), -1)

    chk = 2 if op == 0 else 3
    chnl = get_chnl(argv[(chk - 1):])

    num, entries, privs = _lan_get_ciphers(self, chnl)

    valid_entries = entries[1:num+1]
    if cipher not in valid_entries:
        raise PyCmdsExcept('Cipher Suite ' + argv[0] + ' is not supported by BMC.')

    idx = valid_entries.index(cipher) + 1

    if idx & 1:
        x = int((idx + 1) / 2)    
        if op == 1:     # enable
            privs[x] = (privs[x] & 0xf0) + priv   # change lower nibble
        else:   # disable
            privs[x] &= 0xf0
    else:
        x = int(idx / 2)
        if op == 1:     # enable
            privs[x] = (priv << 4) + (privs[x] & 0x0f)  # change upper nibble
        else:   # disable
            privs[x] &= 0x0f

    # Set LAN Config, parameter #24
    self.intf.issue_cmd(SetLanConfig, chnl, 24, bytes(privs))

def _lan_enable_auth(self, argv, op=1):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 2)

    auth = AUTH_TYPES.get(argv[0], -1)
    if auth < 0:
        raise PyCmdsExcept('Invalid auth type: ' + argv[0], 2)

    priv = str2int(argv[1])
    if priv < 1 or priv > 5:
        raise PyCmdsExcept('Invalid privilege level: ' + argv[1], 2)

    chnl = get_chnl(argv[2:])

    # Get LAN Config, parameter #1
    rsp = self.intf.issue_cmd(GetLanConfig, chnl, 1, 0, 0)
    if not rsp:
        raise PyCmdsExcept('Failed to get the supported auth types.', -1)
    auths = int.from_bytes(rsp, 'little') 
    auth_types = _conv_auth_types(auths)

    if argv[0] not in auth_types:
        raise PyCmdsExcept('Auth type: ' + argv[0] + ' is not supported by BMC.')

    # Get LAN Config, parameter #2
    rsp = self.intf.issue_cmd(GetLanConfig, chnl, 2, 0, 0)
    if not rsp:
        raise PyCmdsExcept('Failed to get the current enabled auth types.', -1)

    rsp = list(rsp)
    rsp = [x & auths for x in rsp]
    
    # Set LAN Config, parameter #2
    if op == 1:
        rsp[priv - 1] |= auth
    else:
        rsp[priv - 1] &= ~auth

    self.intf.issue_cmd(SetLanConfig, chnl, 2, bytes(rsp))

def _lan_enable_imp(self, argv, op=1):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1)

    if argv[1] == 'cipher':
        _lan_enable_cipher(self, argv[2:], op)
    elif argv[1] == 'auth':
        _lan_enable_auth(self, argv[2:], op)
    else:
        raise PyCmdsArgsExcept(2, 1, argv[1])

def _lan_enable(self, argv):
    _lan_enable_imp(self, argv)

def _lan_disable(self, argv):
    _lan_enable_imp(self, argv, 0)

def help_lan(self, argv=None, context=0):
    if context == 0:
        self.print('LAN Commands:')
        self.print('    help')
        self.print('    print [<channel number>]')
        self.print('    alert [<channel number>]')
    if context in (0, 1):
        self.print('    enable cipher <cipher suite> <priv> [<channel number>]')
        self.print('    disable cipher <cipher suite> [<channel number>]')
        self.print('        e.g. enable cipher 2 4')
        self.print('        e.g. disable cipher 2')
    if context in (0, 2):    
        self.print('    enable auth <auth type> <priv> [<channel number>]')
        self.print('    disable auth <auth type> <priv> [<channel number>]')
        self.print('        e.g. enable auth md5 4')
        self.print('        e.g. disable auth md5 4')
        self.print('        Auth Types:')
        for i in AUTH_TYPES.keys():
            self.print(' ' * 9 + '* {0}'.format(i))
        self.print('        Privilege Levels:')
        for i in range(1, 6):
            self.print(' ' * 9 + '* {0}h = {1}'.format(i, USER_PRIVS[i]))

LAN_CMDS = {
    'print': _lan_print,
    'alert': _lan_alert,
    'enable': _lan_enable,
    'disable': _lan_disable,
    'help': help_lan,
}

def do_lan(self, argv):
    do_command(self, argv, help_lan, LAN_CMDS)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()
with open('requirements.txt', 'r') as fh:
    requirements = [line.strip() for line in fh]

setuptools.setup(
    name='hyve-pyipmi',
    version='1.2.1',
    keywords=['ipmi', 'rmcp', 'rmcpp', 'kcs', 'openipmi'],
    author='Janny Au',
    author_email='jannya@hyvedesignsolutions.com',
    description='Pure Python-based IPMI client developed by Hyve Design Solutions.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hyvedesignsolutions/hyve-pyipmi',
    license='BSD 3-Clause License',
    platforms=['linux', 'windows'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    packages=[
        'pyipmi',
        'pyipmi.cmds',
        'pyipmi.intf',
        'pyipmi.intf.ioctl',
        'pyipmi.mesg',
        'pyipmi.util',
    ],
    package_dir={
        'pyipmi': 'src/pyipmi',
        'pyipmi.cmds': 'src/pyipmi/cmds',
        'pyipmi.intf': 'src/pyipmi/intf',
        'pyipmi.intf.ioctl': 'src/pyipmi/intf/ioctl',
        'pyipmi.mesg': 'src/pyipmi/mesg',
        'pyipmi.util': 'src/pyipmi/util',
    },
    scripts=[
        'pyipmi',
        'pyipmr',
        'pyping',
        'pysh',
    ],
)




#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, struct
from .. util.exception import PyIntfExcept, PyIntfCCExcept
from . _rmcpp_msg import IPMI20_Message
from . _crypto import RAKP_NONE, cal_auth_code, map_auth2inte, cal_inte_check

class RAKP_Message:
    def __init__(self, rcsid):
        self.rcsid = rcsid
        self.payload_type = 0
        self.msg_tag = os.urandom(1)

    def pack(self):
        pass

    def unpack(self, payload_type, rsp, exp_len=8):
        if len(rsp) < 8:
            raise PyIntfExcept('Wrong response payload length.')

        tag, cc, _, sid = struct.unpack('<1sBH4s', rsp[:8])

        if payload_type != self.payload_type + 1:
            raise PyIntfExcept('Payload Type is 0x{0:02x}, but expected to be 0x{1:02x}.'.format(
                                                                payload_type, self.payload_type + 1))
        if tag != self.msg_tag:  # Message Tag
            #raise PyIntfExcept('Message Tag does not match.')
            # Workaround for OpenBMC
            pass
        if cc != 0:  # RMCP+ Status Code
            raise PyIntfCCExcept(rsp[1])
        if sid != self.rcsid:
            raise PyIntfExcept('Remote console session ID does not match.')

        if len(rsp) < exp_len:
            raise PyIntfExcept('Wrong response payload length.')

        return rsp[8:]

class RMCPP_OpenSess(RAKP_Message):
    def __init__(self, rcsid, cipher, priv):
        super(RMCPP_OpenSess, self).__init__(rcsid)
        self.cipher = cipher  # (auth, inte, conf)
        self.priv = priv
        self.payload_type = 0x10

    def pack(self):
        # RMCP+ Open Session Request, payload_type = 0x10
        # console -> BMC
        
        auth = struct.pack('<BHBB3s', 0, 0, 8, self.cipher[0], b'\0')
        inte = struct.pack('<BHBB3s', 1, 0, 8, self.cipher[1], b'\0')
        conf = struct.pack('<BHBB3s', 2, 0, 8, self.cipher[2], b'\0')

        payload = struct.pack('<1sBH4s', self.msg_tag, self.priv, 0, self.rcsid)
        payload += auth + inte + conf

        return payload

    def unpack(self, payload_type, rsp):
        # RMCP+ Open Session Response
        # BMC -> console, payload_type = 0x11
        payload = super(RMCPP_OpenSess, self).unpack(payload_type, rsp, 36)
        mssid, auth, inte, conf = struct.unpack('<4s8s8s8s', payload)
        algo = lambda x: x[4] & 0x3f

        return (mssid, (algo(auth), algo(inte), algo(conf)))

class RAKP_1_2(RAKP_Message):
    def __init__(self, rcsid, mssid, rcrn, priv, user=None, passwd=None, auth=RAKP_NONE):
        super(RAKP_1_2, self).__init__(rcsid)
        self.mssid = mssid
        self.rcrn = rcrn
        self.priv = priv
        self.user = user
        self.passwd = passwd
        self.auth = auth
        self.payload_type = 0x12

    def pack(self):
        # RAKP 1: console -> BMC
        if self.user is not None:
            name_fmt = '<1s3s4s16sBHB' + str(len(self.user)) + 's'
            payload = struct.pack(name_fmt, self.msg_tag, b'\0', self.mssid,
                                  self.rcrn, self.priv, 0, len(self.user), self.user)
        else:
            payload = struct.pack('<1s3s4s16sBHB', self.msg_tag, b'\0', self.mssid,
                                  self.rcrn, self.priv, 0, 0)

        return payload

    def unpack(self, payload_type, rsp):
        # RAKP 2 : BMC -> console
        payload = super(RAKP_1_2, self).unpack(payload_type, rsp, 40)

        msrn = payload[:16]
        msguid = payload[16:32]
        auth_code = payload[32:]

        # verify auth code
        # auth_code = H(rcsid, mssid, rcrn, msrn, msguid, priv, len(user), user)
        # key = user password
        if self.auth != RAKP_NONE and self.passwd is not None:
            if len(auth_code) == 0:
                raise PyIntfExcept('Key Exchange Authentication Code is not valid in RAKP 2.')           

            if self.user is not None:
                data = struct.pack('<4s4s16s16s16sBB', self.rcsid, self.mssid, self.rcrn, msrn, msguid,
                                   self.priv, len(self.user))
                data += self.user
            else:
                data = struct.pack('<4s4s16s16s16sBB', self.rcsid, self.mssid, self.rcrn, msrn, msguid,
                                   self.priv, 0)

            auth_code_expected = cal_auth_code(self.auth, self.passwd, data)

            if auth_code != auth_code_expected:
                raise PyIntfExcept('Key Exchange Authentication Code is not valid in RAKP 2.')           

        return (msrn, msguid)

class RAKP_3_4(RAKP_Message):
    def __init__(self, rcsid, mssid, msrn, rcrn, msguid, 
                 priv, user=None, passwd=None, auth=RAKP_NONE, sik=None):
        super(RAKP_3_4, self).__init__(rcsid)        
        self.mssid = mssid
        self.msrn = msrn
        self.rcrn = rcrn
        self.msguid = msguid
        self.priv = priv
        self.user = user
        self.passwd = passwd
        self.auth = auth
        self.sik = sik
        self.payload_type = 0x14

    def pack(self):
        # RAKP 3: console -> BMC
        payload = struct.pack('<1s3s4s', self.msg_tag, b'\0', self.mssid)

        # calculate auth code
        # auth_code = H(msrn, rcsid, priv, len(user), user)
        # key = user password
        if self.auth != RAKP_NONE and self.passwd is not None:
            if self.user != None:
                data = struct.pack('<16s4sBB', self.msrn, self.rcsid, 
                                   self.priv, len(self.user))
                data += self.user
            else:
                data = struct.pack('<16s4sBB', self.msrn, self.rcsid, 
                                   self.priv, 0)
        
            auth_code = cal_auth_code(self.auth, self.passwd, data)
            payload += auth_code

        return payload

    def unpack(self, payload_type, rsp):
        # RAKP 4: BMC -> console
        auth_code = super(RAKP_3_4, self).unpack(payload_type, rsp, 8)

        # verify auth code
        # auth_code = H(rcrn, mssid, msguid)
        # key = SIK
        if self.auth != RAKP_NONE and self.sik is not None:
            if len(auth_code) == 0:
                raise PyIntfExcept('Key Exchange Authentication Code is not valid in RAKP 4.')           

            data = struct.pack('16s4s16s', self.rcrn, self.mssid, self.msguid)
            inte = map_auth2inte(self.auth)
            auth_code_expected = cal_inte_check(inte, self.sik, data)

            if auth_code != auth_code_expected:
                raise PyIntfExcept('Key Exchange Authentication Code is not valid in RAKP 4.')           



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
# The base class of all the exceptions defined in PyIPMI
class PyExcept(Exception):
    def __str__(self):
        return type(self).__name__ + ': ' + Exception.__str__(self)

# The exceptions in RMCP or RMCP+ protocols
class PyIntfExcept(PyExcept):
    pass 

# The exceptions when the response doesn't match the request
# by the sequence number
class PyIntfSeqExcept(PyIntfExcept):
    pass 

# The exceptions of reading config
class PyConfExcept(PyExcept):
    pass

# The exceptions for RMCP+ and RAKP Messages
class PyIntfCCExcept(PyIntfExcept):
    def __init__(self, cc):
        self.cc = cc

    def __str__(self):
        ERR_CCS = (
            'No errors',
            'Insufficient resources to create a session',
            'Invalid Session ID',
            'Invalid payload type',
            'Invalid authentication algorithm',
            'Invalid integrity algorithm',
            'No matching authentication payload',
            'No matching integrity payload',
            'Inactive Session ID',
            'Invalid role',
            'Unauthorized role or privilege level requested',
            'Insufficient resources to create a session at the requested role',
            'Invalid name length',
            'Unauthorized name',
            'Unauthorized GUID (RAKP2 BMC to remote console)',
            'Invalid integrity check value',
            'Invalid confidentiality algorithm',
            'No Cipher Suite match with propsed security algorithms',
            'Illegal or unrecognized parameter',
        )

        if self.cc <= 0x12:  msg = ERR_CCS[self.cc]
        else:  msg = 'Reserved'

        return 'Failed to establish an RMCP+ session.  RMCP+ Status Code = {0:02x}h ({1}.)'.format(self.cc, msg)

# The exceptions in IPMI commands
class PyMesgExcept(PyExcept):
    pass 

# The exceptions in IPMI responses when 
# the completion code is non-zero
class PyMesgCCExcept(PyMesgExcept):
    def __init__(self, netfn, cmd, cc):
        self.netfn = netfn
        self.cmd = cmd
        self.cc = cc

    def __str__(self):
        ERR_CCS = {
            0xc0: 'Node Busy.', 
            0xc1: 'Invalid Command.',
            0xc2: 'Invalid LUN.',
            0xc3: 'Timeout while processing command.',
            0xc4: 'Out of space.',
            0xc5: 'Invalid Reservation ID.',
            0xc6: 'Request data truncated.',
            0xc7: 'Request data length invalid.',
            0xc8: 'Request data field length limit exceeded.',
            0xc9: 'Parameter out of range.',
            0xca: 'Cannot return number of requested data bytes.',
            0xcb: 'Requested Sensor, data, or record not present.',
            0xcc: 'Invalid data field in Request.',
            0xcd: 'Command illegal for specified sensor or record type.',
            0xce: 'Command response could not be provided.',
            0xcf: 'Cannot execute duplicated request.',
            0xd0: 'Command response could not be provided. SDR Repository in update mode.',
            0xd1: 'Command response could not be provided. Device in firmware update mode.',
            0xd2: 'Command response could not be provided. BMC init or init agent in progress.',
            0xd3: 'Destination unavailable',
            0xd4: 'Cannot execute command due to insufficient privilege level or other security-based restriction.',
            0xd5: 'Cannot execute command. Command or request parameters not supported in present state.',
            0xd6: 'Cannot execute command. Parameter is illegal because command sub-functions has been disabled or is unavailable.',
            0xff: 'Unspecified error.',
        }

        return ('Command failed: NetFn={0:02X}h, CMD={1:02X}h, CC={2:02X}h ({3})'
                .format(self.netfn, self.cmd, self.cc, 
                        ERR_CCS.get(self.cc, 'Unknown Completion Code.')))

# The exceptions in PyIPMI commands
class PyCmdsExcept(PyExcept):
    def __init__(self, msg, context=0):
        self.context = context
        super(PyCmdsExcept, self).__init__(msg)

    def __str__(self):
        return Exception.__str__(self)

class PyCmdsArgsExcept(PyCmdsExcept):
    def __init__(self, opt, context=0, arg=None):
        if opt == 1:
            super(PyCmdsArgsExcept, self).__init__('Insufficient arguments.', context)
        elif opt == 2:
            super(PyCmdsArgsExcept, self).__init__('Unknown argument: {0}.'.format(arg), context)
        elif opt == 3:
            super(PyCmdsArgsExcept, self).__init__('Invalid input argument: {0}.'.format(arg), context)
        elif opt == 4:
            msg = 'Unexpected sensor type: {0}'.format(arg)
            msg += '. Please specify the sensor type by an integer.\n'
            msg += 'Use \'list type\' to get the full list of all the sensor types.'
            super(PyCmdsArgsExcept, self).__init__(msg, context)
        else:
            super(PyCmdsArgsExcept, self).__init__('Unspecified error occurred.', context)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import cmd, sys
from .. util import PyTest

class PyShell(cmd.Cmd, PyTest):
    intro = 'Welcome to the PyIPMI shell.  Type help or ? to list commands.\n'
    prompt = 'pysh> '
    doc_header = 'PyIPMI shell commands (type help <topic>):'
    file = None

    def __init__(self, print_file=sys.stdout):
        PyTest.__init__(self, None, False, True)
        cmd.Cmd.__init__(self)
        self.print_file = print_file

    from ._lan import do_lan, help_lan
    from ._mc import do_mc, help_mc
    from ._chassis import do_chassis, help_chassis
    from ._power import do_power, help_power
    from ._sdr import do_sdr, help_sdr
    from ._sel import do_sel, help_sel
    from ._sensor import do_sensor, help_sensor
    from ._user import do_user, help_user
    from ._fru import do_fru, help_fru
    from ._raw import do_raw, help_raw

    from ._common import print
    from ._common import print_rsp
    
    def do_exit(self, arg):
        self.intf.__del__()
        return True
    
    def help_exit(self):
        print('Type \'exit\' to leave the PyIPMI Shell.')

    def do_selftest(self, argv):
        print('Calling raw 6 1...')
        self.do_raw(['6', '1'])
        print('\nCalling mc info...')
        self.do_mc(['info'])
        print('\nCalling chassis status...')
        self.do_chassis(['status'])
        print('\nCalling power status...')
        self.do_power(['status'])
        print('\nCalling sdr list...')
        self.do_sdr(['list'])
        print('\nCalling sel list...')
        self.do_sel(['list'])
        print('\nCalling sensor list...')
        self.do_sensor(['list'])
        print('\nCalling lan print 1...')
        self.do_lan(['print', '1'])
        print('\nCalling user list...')
        self.do_user(['list'])
        print('\nCalling fru print...')
        self.do_fru(['print'])
        print('\nselftest done.')
    
    def help_selftest(self):
        print('Type \'selftest\' to perform a self test.')

    def precmd(self, line):
        text = line.lower()
        if len(text) > 1 and text[0] == '[' and text[-1] == ']':
            # raw command
            text = 'raw ' + text[1:-1]

        return text

    def get_pos(self, text, line, begidx, endidx):
        count = 0
        c_prev = ''
        for c in line:
            if c == begidx:  break
            if c == ' ' and c_prev != ' ':
                count += 1

            c_prev = c

        args = line.split(' ')

        if count > 1:  ret = (count, args[count-1], args[count-2])
        elif count == 1:  ret = (count, args[count-1], None)
        else:  ret = (0, None)

        return ret
        
    def match_words(self, text, words):
        ret = []
        for word in words:
            if text == word[:len(text)]:
                ret.append(word)

        return ret if ret else None

    def complete_chassis(self, text, line, begidx, endidx):
        pos, arg_prev, arg_1st = self.get_pos(text, line, begidx, endidx)
        words_d = {
            'power': ('status', 'on', 'off', 'cycle', 'reset', 'diag', 'soft',),
            'policy': ('list', 'always-on', 'previous', 'always-off',),
            'identify': ('on', 'off',),
            'bootdev': ('none', 'pxe', 'safe', 'diag', 'cdrom', 'bios', 'floppy',),
        }

        if pos == 1:
            words = ('bootdev', 'identify', 'restart_cause', 'status', 
                     'power', 'policy', 'poh',)
            
            return self.match_words(text, words)
        
        if pos == 2:
            words = words_d.get(arg_prev, None)

            if words is None:   return None
            return self.match_words(text, words)

        if pos == 3:
            words = ('clear-cmos', )
            if arg_1st == 'bootdev':
                return self.match_words(text, words)
            else:  
                return None
        
        return None

    def complete_power(self, text, line, begidx, endidx):
        pos, *_ = self.get_pos(text, line, begidx, endidx)
        if pos == 1:
            words = ('status', 'on', 'off', 'cycle', 'reset', 'diag', 'soft',)
            return self.match_words(text, words)

        return None

    def complete_mc(self, text, line, begidx, endidx):
        pos, arg_prev, arg_1st = self.get_pos(text, line, begidx, endidx)

        if pos == 1:
            words = ('help', 'reset', 'guid', 'info', 'selftest', 'watchdog',
                     'getenables', 'setenables', 'getsysinfo', 'setsysinfo',)
            return self.match_words(text, words)

        if pos == 2:
            if arg_prev == 'watchdog':
                words = ('get', 'reset')
                return self.match_words(text, words)
            elif arg_prev == 'setenables':
                words = ('recv_msg_intr', 'event_msg_intr', 'event_msg', 'system_event_log',
                         'oem0', 'oem1', 'oem2',)
                return self.match_words(text, words)
            elif arg_prev in ('getsysinfo', 'setsysinfo'):
                words = ('system_fw_version', 'system_name', 'primary_os_name', 'os_name',)
                return self.match_words(text, words)
            else:
                return None

        if pos == 3:
            if arg_1st == 'setenables':
                words = ('on', 'off')
                return self.match_words(text, words)
            else:
                return None

        return None

    def complete_sdr(self, text, line, begidx, endidx):
        pos, arg_prev, _ = self.get_pos(text, line, begidx, endidx)
        words_list = ['list', 'elist', 'slist', 'vlist',]

        if pos == 1:
            words = words_list + ['info', 'help']
            return self.match_words(text, words)

        if pos == 2:
            if arg_prev in words_list:
                words = ('all', 'full', 'compact', 'event', 'fru', 'mcloc', 'type',)
                return self.match_words(text, words)
            else:
                return None

        return None

    def complete_sel(self, text, line, begidx, endidx):
        pos, *_ = self.get_pos(text, line, begidx, endidx)

        if pos == 1:
            words = ('info', 'list', 'elist', 'vlist', 'clear', 'time', 'help',)
            return self.match_words(text, words)

        return None
    
    def complete_sensor(self, text, line, begidx, endidx):
        pos, arg_prev, _ = self.get_pos(text, line, begidx, endidx)
        words_list = ['list', 'vlist',]

        if pos == 1:
            words = words_list + ['help',]
            return self.match_words(text, words)

        if pos == 2:
            if arg_prev in words_list:
                words = ('type',)
                return self.match_words(text, words)
            else:
                return None

        return None

    def complete_user(self, text, line, begidx, endidx):
        pos, arg_prev, _ = self.get_pos(text, line, begidx, endidx)

        if pos == 1:
            words = ('info', 'list', 'set', 'test', 'enable', 'disable',)
            return self.match_words(text, words)

        if pos == 2:
            if arg_prev == 'set':
                words = ('name', 'password', 'priv',)
                return self.match_words(text, words)
            else:
                return None

        return None

    def complete_lan(self, text, line, begidx, endidx):
        pos, arg_prev, _ = self.get_pos(text, line, begidx, endidx)

        if pos == 1:
            words = ('help', 'print', 'enable', 'disable', 'alert',)
            return self.match_words(text, words)
        if pos == 2:
            words = ('auth', 'cipher',)
            if arg_prev in ('enable', 'disable'):
                return self.match_words(text, words)
            else:
                return None

        return None

    def complete_fru(self, text, line, begidx, endidx):
        pos, *_ = self.get_pos(text, line, begidx, endidx)

        if pos == 1:
            words = ('print', 'read', 'write',)
            return self.match_words(text, words)

        return None




#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os
from .. mesg.ipmi_storage import GetSDRRepoInfo, GetSDRRepoAllocInfo, RevSDR, \
                                ClearSDR, AddSDR, PartialAddSDR
from . _common import get_sensor_readings, print_sensor_list1, print_sensor_list2, \
                    print_sensor_list3, print_sensor_list4, conv_time, do_command, \
                    bcd2str, str2int, get_sdr_repo, conv_sensor_type, conv_entity_id, \
                    print_sensor_type_list, get_one_sdr
from . _consts import TupleExt, ENTITY_DEVICE_CODES
from .. util.exception import PyCmdsArgsExcept, PyCmdsExcept

def _get_event_sdr(self, opt):
    sdr_repo = get_sdr_repo(self)

    for sdr1 in sdr_repo[3]:
        sensor_num = sdr1[2]
        entity_id = sdr1[3]
        entity_inst = sdr1[4] & 0x7f
        sensor_name = sdr1[12:].decode('latin_1')

        if opt in (1, 2, 3):
            yield (sensor_num, entity_id, entity_inst, sensor_name)
        else:
            sensor_type_id = sdr1[5]
            sensor_type = conv_sensor_type(sensor_type_id)
            sensor_type += ' ({0:02X}h)'.format(sensor_type_id)

            entity_str = conv_entity_id(entity_id)
            yield (sensor_num, entity_id, entity_inst, sensor_name, sensor_type, entity_str)


def _get_fru_sdr(self, opt):
    sdr_repo = get_sdr_repo(self)

    for sdr1 in sdr_repo[0x11]:
        if sdr1[2] & 0x80 == 0:
            fru_type = 'Physical FRU'  
            fru_id = '{0}:{1:02x}h'.format(sdr1[2] & 7, sdr1[1])
        else:
            fru_type = 'Logical FRU'
            fru_id = '{0:02x}h'.format(sdr1[1])

        entity_id = sdr1[7]
        entity_inst = sdr1[8] 
        fru_name = sdr1[11:].decode('latin_1')

        if opt in (1, 2, 3):
            yield (fru_name, entity_id, entity_inst, fru_type, fru_id)
        else:
            chnl = sdr1[3]
            dev_type = sdr1[5]
            dev_mod = sdr1[6]
            dev_acc_addr = sdr1[0]
            if dev_type == 0x10:  dev_type_idx = (dev_type << 8) + dev_mod
            else: dev_type_idx = dev_type

            entity_str = conv_entity_id(entity_id)
            dev_type_str = ENTITY_DEVICE_CODES.get(dev_type_idx, '')

            yield (fru_name, entity_id, entity_inst, fru_type, fru_id, chnl,
                   entity_str, dev_type_str, dev_type, dev_mod, dev_acc_addr)

def _get_mcloc_sdr(self, opt):
    sdr_repo = get_sdr_repo(self)

    for sdr1 in sdr_repo[0x12]:
        mc_type = 'Dynamic MC' if sdr1[2] & 0x20 == 0 else 'Static MC'
        mc_id = '{0:02X}h'.format(sdr1[0])

        entity_id = sdr1[7]
        entity_inst = sdr1[8] 
        mc_name = sdr1[11:].decode('latin_1')

        if opt in (1, 2, 3):            
            yield (mc_name, entity_id, entity_inst, mc_type, mc_id)
        else:
            entity_str = conv_entity_id(entity_id)
            chnl = sdr1[1]
            acpi = ('Not Required', 'Required')
            acpi_sys = acpi[(sdr1[2] & 0x80) >> 7]
            acpi_dev = acpi[(sdr1[2] & 0x40) >> 6]
            log_init_agent = ('No', 'Yes')[(sdr1[2] & 8) >> 3]            
            event_gen = ('Enabled', 'Disabled', 'Do Not Init Controller', '')[sdr1[2] & 3]

            caps = sdr1[3]
            list_caps = []
            cap_names = ('Sensor Device', 'SDR Repository Device', 'SEL Device', 
                'FRU Inventory Device', 'IPMB Event Receiver', 'IPMB Event Generator',
                'Bridge', 'Chassis Device')

            idx = 0
            while caps:
                if caps & 1: list_caps.append(cap_names[idx])
                caps >>= 1 
                idx += 1

            yield (mc_name, entity_id, entity_inst, mc_type, mc_id, entity_str, chnl,
                   acpi_sys, acpi_dev, log_init_agent, event_gen, list_caps)

def _sdr_list_event(self, opt=1, hdl=None):
    if opt in (1, 3):
        for num, *_, name in _get_event_sdr(self, opt):
            self.print('{0:02X}h | {1:16} | Event-Only'.format(num, name))
    elif opt == 2:
        for num, entity_id, entity_inst, name in _get_event_sdr(self, opt):
            self.print('{0:02X}h | {1:16} | {2:>2}.{3:02} | Event-Only'.format( 
                  num, name, entity_id, entity_inst))
    else:
        for sensor_num, entity_id, entity_inst, sensor_name, sensor_type,\
            entity_str in _get_event_sdr(self, opt):

            self.print('Sensor ID        : {0} ({1:02X}h)'.format(sensor_name, sensor_num))
            self.print('Entity ID        : {0}.{1}'.format(entity_id, entity_inst), end='')

            if entity_str != '': self.print(' (' + entity_str + ')')
            else: self.print('')
            self.print('Sensor Type      : ' + sensor_type)
            self.print('')

def _sdr_list_fru(self, opt=1, hdl=None):
    if opt in (1, 3):
        for fru_name, _, _, fru_type, fru_id in _get_fru_sdr(self, opt):
            self.print('--- | {0:16} | {1} @ {2}'.format( 
                  fru_name, fru_type, fru_id))
    elif opt == 2:
        for fru_name, entity_id, entity_inst, fru_type, fru_id in _get_fru_sdr(self, opt):
            self.print('--- | {0:16} | {1:>2}.{2:02} | {3} @ {4}'.format( 
                  fru_name, entity_id, entity_inst, fru_type, fru_id))
    else:
        for fru_name, entity_id, entity_inst, fru_type, fru_id, chnl, entity_str,\
            dev_type_str, dev_type, dev_mod, dev_acc_addr in _get_fru_sdr(self, opt):

            self.print('Device ID               : ' + fru_name)
            self.print('Entity ID               : {0}.{1}'.format(entity_id, entity_inst), end='')

            if entity_str != '': self.print(' (' + entity_str + ')')
            else: self.print('')

            self.print('Device Access Address   : {0:02X}h'.format(dev_acc_addr))
            self.print(fru_type + ' Device      : ' + fru_id)
            self.print('Channel Number          : {0}'.format(chnl))            
            
            self.print('Device Type             : {0:02X}h.{1:02X}h'.format(dev_type, dev_mod), end='')
            if dev_type != '': self.print(' (' + dev_type_str + ')')
            else: self.print('')

            self.print('')

def _sdr_list_mcloc(self, opt=1, hdl=None):
    if opt in (1, 3):
        for mc_name, _, _, mc_type, mc_id in _get_mcloc_sdr(self, opt):
            self.print('--- | {0:16} | {1} @ {2}'.format(
                  mc_name, mc_type, mc_id))
    elif opt == 2:
        for mc_name, entity_id, entity_inst, mc_type, mc_id in _get_mcloc_sdr(self, opt):
            self.print('--- | {0:16} | {1:>2}.{2:02} | {3} @ {4}'.format(
                  mc_name, entity_id, entity_inst, mc_type, mc_id))
    else:
        for mc_name, entity_id, entity_inst, mc_type, mc_id, entity_str, chnl, acpi_sys,\
            acpi_dev, log_init_agent, event_gen, list_caps in _get_mcloc_sdr(self, opt):

            self.print('Device ID               : ' + mc_name)
            self.print('Entity ID               : {0}.{1}'.format(entity_id, entity_inst), end='')

            if entity_str != '': self.print(' (' + entity_str + ')')
            self.print('Device Slave Address    : ' + mc_id)
            self.print('Channel Number          : {0}'.format(chnl))
            self.print('ACPI System P/S Notif   : ' + acpi_sys)
            self.print('ACPI Device P/S Notif   : ' + acpi_dev)
            self.print('Controller Presence     : ' + mc_type)
            self.print('Logs Init Agent Errors  : ' + log_init_agent)
            self.print('Event Message Gen       : ' + event_gen)

            self.print('Device Capabilities     : ', end='')
            if not list_caps:
                self.print('none')
            else:
                self.print('[{0}]'.format(list_caps[0]))
                for i in list_caps[1:]:
                    self.print(' ' * 26 + '[{0}]'.format(i))            
            self.print('')

def _sdr_list_all(self, opt=1, print_hdl=None):
    # SDR type 1h, 2h
    reading_all = get_sensor_readings(self, opt)
    print_hdl(self, reading_all)

    # Other SDR types
    for key in SDR_PRINT_HDL.keys():
        if key == 'all':    continue
        SDR_PRINT_HDL[key](self, opt, print_hdl)

def _sdr_list(self, argv):
    filter_sdr = 0
    filter_sensor_type = 0
    opt, print_hdl = SENSOR_PRINT_HDL[argv[0]]    

    if len(argv) > 1:
        if argv[1] == 'full':
            filter_sdr = 1
        elif argv[1] == 'compact':
            filter_sdr = 2
        elif argv[1] == 'type':
            if len(argv) > 2:
                filter_sensor_type = str2int(argv[2])
                if filter_sensor_type < 0:
                    raise PyCmdsArgsExcept(4, 0, argv[2])
            else:
                print_sensor_type_list(self)
                return
        else:
            if argv[1] in SDR_PRINT_HDL.keys():
                return SDR_PRINT_HDL[argv[1]](self, opt, print_hdl)
            else:
                raise PyCmdsArgsExcept(2, 0, argv[1])

    reading_all = get_sensor_readings(self, opt, filter_sdr, filter_sensor_type)
    print_hdl(self, reading_all)

def _sdr_info(self, argv):
    ver, rec, free, add_ts, erase_ts, op = self.intf.issue_cmd(GetSDRRepoInfo)
    self.print('SDR Version                         : ' + bcd2str(ver, True))
    self.print('Record Count                        : {0}'.format(rec))
    self.print('Free Space                          : {0} bytes'.format(free))
    self.print('Most Recent Addition                : ' + conv_time(add_ts))
    self.print('Most Recent Erase                   : ' + conv_time(erase_ts))

    UC = 'unspecified'
    update = (op & 0x60) >> 5
    sdr_update_support = (UC, 'non-modal', 'modal', 'both')[update]

    get_op_support = lambda mask: 'yes' if op & mask else 'no'

    self.print('SDR Overflow                        : ' + get_op_support(0x80))
    self.print('SDR Repository Update Supported     : ' + sdr_update_support)
    self.print('Delete SDR Supported                : ' + get_op_support(8))
    self.print('Partial Add SDR Supported           : ' + get_op_support(4))
    self.print('Reserve SDR Repository Supported    : ' + get_op_support(2))
    self.print('SDR Repository Alloc Info Supported : ' + get_op_support(1))    

    if op & 1:
        t1 = self.intf.issue_cmd(GetSDRRepoAllocInfo)

        alloc_units = '{0}'.format(t1[0]) if t1[0] != 0 else UC
        alloc_unit_size = '{0} bytes'.format(t1[1]) if t1[1] != 0 else UC

        self.print('# of Alloc Units                    : ' + alloc_units)
        self.print('Alloc Unit Size                     : ' + alloc_unit_size)
        self.print('# Free Units                        : {0}'.format(t1[2]))
        self.print('Largest Free Blk                    : {0}'.format(t1[3]))
        self.print('Max Record Size                     : {0}'.format(t1[4]))    

def _sdr_get(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1)

    sensor_num = str2int(argv[1])
    if sensor_num == -1:
        raise PyCmdsArgsExcept(3, 0, argv[1])

    opt, print_hdl = SENSOR_PRINT_HDL[argv[0]]
    reading_all = get_sensor_readings(self, opt, 0, 0, False, sensor_num)
    print_hdl(self, reading_all)

def _sdr_dump(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1)

    next_id = b'\0'
    sdr_count = 0

    _, rec, *_ = self.intf.issue_cmd(GetSDRRepoInfo)
    if rec == 0:  
        raise PyCmdsExcept('SDR repository is empty.', -1)

    resv_id, = self.intf.issue_cmd(RevSDR)
    with open(argv[1], 'wb') as out_file:
        self.print('Dumping Sensor Data Repository to \'' + argv[1] + '\'.')
        while next_id != b'\xff\xff':
            if sdr_count > rec: break   # already got more SDRs than the total # 
            sdr1 = get_one_sdr(self, resv_id, next_id)
            next_id = sdr1[:2]
            out_file.write(sdr1[2:])
            sdr_count += 1

def _sdr_fill(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1)

    file_size = os.stat(argv[1]).st_size

    # clear SDR repo
    resv_id, = self.intf.issue_cmd(RevSDR)
    self.intf.issue_cmd(ClearSDR, resv_id, 0xaa)
    self.print('The SDR repo has been cleared.')

    # adding SDR
    self.print('Filling Sensor Data Repository from \'' + argv[1] + '\'.')
    with open(argv[1], 'rb') as in_file:
        count = 0
        while count < file_size:
            sdr1 = in_file.read(5)
            count += 5
            rec_len = sdr1[4]
            sdr1 += in_file.read(rec_len)
            count += rec_len

            wlen = len(sdr1)
            if wlen <= 32:
                # directly call Add SDR
                self.intf.issue_cmd(AddSDR, sdr1)
            else:
                # reserve SDR
                resv_id, = self.intf.issue_cmd(RevSDR)
                # partial add SDR
                offset = 0
                rec_id, = self.intf.issue_cmd(PartialAddSDR, resv_id, b'\0', offset, 
                                              0, sdr1[:32])
                wlen -= 32
                offset += 32
                while wlen > 0:
                    if wlen <= 32:
                        # the last request
                        self.intf.issue_cmd(PartialAddSDR, resv_id, rec_id, offset, 
                                            1, sdr1[offset:])
                        wlen = 0
                    else:
                        # adding in progress
                        self.intf.issue_cmd(PartialAddSDR, resv_id, rec_id, offset, 
                                            0, sdr1[offset:offset+32])
                        offset += 32
                        wlen -= 32

def help_sdr(self, argv=None, context=0):
    self.print('SDR Commands:')
    self.print('    list | elist | slist | vlist [options]')
    self.print('        all')
    self.print('        full')
    self.print('        compact')
    self.print('        event')
    self.print('        fru')
    self.print('        mcloc')
    self.print('        type [<sensor_type>]')
    self.print('    get <sensor number>')
    self.print('    dump <output file>')
    self.print('    fill <input file>')
    self.print('    info')
    self.print('    help')

SDR_CMDS = {
    'list': _sdr_list,
    'elist': _sdr_list,
    'slist': _sdr_list,
    'vlist': _sdr_list,
    'get': _sdr_get,
    'info': _sdr_info,
    'dump': _sdr_dump,
    'fill': _sdr_fill,
    'help': help_sdr,
}

SENSOR_PRINT_HDL = {
    'list': (1, print_sensor_list1),
    'elist': (2, print_sensor_list2),
    'slist': (3, print_sensor_list3),
    'vlist': (4, print_sensor_list4),
    'get': (4, print_sensor_list4),
}

SDR_PRINT_HDL = {
    'event': _sdr_list_event,
    'fru': _sdr_list_fru,
    'mcloc': _sdr_list_mcloc,
    'all': _sdr_list_all,
}

def do_sdr(self, argv):
    do_command(self, argv, help_sdr, SDR_CMDS)    



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, struct
from .. mesg.ipmi_app import GetDeviceID
from .. mesg.ipmi_storage import GetFruAreaInfo, ReadFru, WriteFru
from .. util import checksum
from .. util.exception import PyCmdsExcept, PyCmdsArgsExcept
from . _consts import TupleExt, CHASSIS_TYPES
from . _common import do_command, get_sdr_repo, str2int, conv_time

FRU_BUF_SIZE = 32
AREA_NAMES = {'c': 'Chassis', 'b': 'Board', 'p': 'Product'}

def _fru_get_common(self, fru_id, area_size):
    if area_size < 8:
        raise PyCmdsExcept('FRU size is too small to have a common header.', -1)

    data = self.intf.issue_cmd(ReadFru, fru_id, 0, 8)
    if data[0] != 8:  
        raise PyCmdsExcept('Failed to get FRU common header.', -1)

    data = data[1:]

    hdr = struct.unpack('B' * 8, data)
    offset_c = hdr[2]
    offset_b = hdr[3]
    offset_p = hdr[4]
    
    return (offset_c, offset_b, offset_p)

def _fru_get_area(self, fru_id, offset, area_size, area_name):
    if area_size < offset * 8 + 2:
        raise PyCmdsExcept('FRU size is too small to have a {0} Info Area.'.format(area_name), -1)

    # Area Format Version, Area Length
    data = self.intf.issue_cmd(ReadFru, fru_id, offset * 8, 2)
    if data[0] != 2:  
        raise PyCmdsExcept('Failed to get {0} Info Area Length.'.format(area_name), -1)

    data = data[1:]
    area_len = data[1]

    # xx Info Area
    read_count = area_len * 8 - 2
    read_offset = offset * 8 + 2    
    ret = b''

    if area_size < read_offset + read_count:
        raise PyCmdsExcept('FRU size is too small to have a {0} Info Area.'.format(area_name), -1)

    while read_count > 0:
        if read_count > 255:
            read_size = 255
        else:
            read_size = read_count

        data = self.intf.issue_cmd(ReadFru, fru_id, read_offset, read_size)
        read_offset += data[0]
        read_count -= data[0]
        ret += data[1:]

    return ret

def _fru_get_field(data, idx):
    field_len = data[idx] & 0x3f
    field_end = idx+field_len+1
    if field_len == 0 or field_end >= len(data):
        return (0, b'') 
    field = data[idx+1:field_end]
    return (field_len, field)

def _fru_print_chassis_type(self, title, chs_type):
    self.print('  {0:24}: {1}'.format(title, CHASSIS_TYPES.get(chs_type, chs_type)))

def _fru_print_field(self, title, field):
    self.print('  {0:24}: {1}'.format(title, field.decode('latin_1')))

def _fru_print_all_fields(self, data, area_name, titles, idx_s):
    idx = idx_s
    count = 1
    start = data[idx]
    while start != 0xc1:
        field_len, field = _fru_get_field(data, idx)

        if field_len > 0:
            _fru_print_field(self, area_name + ' ' + titles.get(count, 'Extra'), field)

        idx += field_len + 1
        count += 1
        if idx >= len(data): return
        start = data[idx]

def _fru_print_chassis(self, fru_id, offset, area_size):
    area_name = AREA_NAMES['c']

    # Retrieve the Chassis Info Area
    data = _fru_get_area(self, fru_id, offset, area_size, area_name)    

    titles = TupleExt(('Type', 'Part Number', 'Serial Number'))
    # Print Chassis Type
    chs_type = data[0]
    _fru_print_chassis_type(self, area_name + ' ' + titles[0], chs_type)

    # Print other fields
    _fru_print_all_fields(self, data, area_name, titles, 1)

def _fru_print_board_time(self, title, mfg_time):
    t1 = struct.unpack('B' * 3, mfg_time)
    mtime = (((t1[2] << 8) + t1[1]) << 8) + t1[0]
    ts = mtime * 60 + 820454400
    self.print('  {0:24}: {1}'.format(title, conv_time(ts)))

def _fru_print_board(self, fru_id, offset, area_size):
    area_name = AREA_NAMES['b']

    # Retrieve the Board Info Area
    data = _fru_get_area(self, fru_id, offset, area_size, area_name)    

    titles = TupleExt(('Mfg. Time', 'Manufacturer', 'Product Name', 'Serial Number',
                       'Part Number', 'FRU File ID',))

    # Print Mfg. Date / Time
    mfg_time = data[1:4]
    _fru_print_board_time(self, area_name + ' ' + titles[0], mfg_time)

    # Print other fields
    _fru_print_all_fields(self, data, area_name, titles, 4)

def _fru_print_product(self, fru_id, offset, area_size):
    area_name = AREA_NAMES['p']

    # Retrieve the Board Info Area
    data = _fru_get_area(self, fru_id, offset, area_size, area_name)    

    titles = TupleExt(('none', 'Manufacturer', 'Name', 'Part Number', 'Version',
                       'Serial Number', 'Asset Tag', 'FRU File ID',))

    # Print product fields
    _fru_print_all_fields(self, data, area_name, titles, 1)

def _fru_print_one(self, fru_id, area_size):
    offset_c, offset_b, offset_p = _fru_get_common(self, fru_id, area_size)
    if offset_c != 0:  _fru_print_chassis(self, fru_id, offset_c, area_size)
    if offset_b != 0:  _fru_print_board(self, fru_id, offset_b, area_size)
    if offset_p != 0:  _fru_print_product(self, fru_id, offset_p, area_size)

def _fru_print(self, argv):
    title = 'FRU Device Description'
    req_id = -1
    if len(argv) > 1:
        req_id = str2int(argv[1])

    # Print BMC FRU
    if req_id < 1:
        t1 = self.intf.issue_cmd(GetDeviceID)
        if t1[5] & 8:  
            # BMC has FRU device
            area_size, _ = self.intf.issue_cmd(GetFruAreaInfo, 0)
            if area_size == 0:  
                raise PyCmdsExcept('The area size of the FRU device with ID 0 is zero.', -1)
            self.print('{0:26}: {1}'.format(title, 'Builtin FRU Device (ID 0)'))
            _fru_print_one(self, 0, area_size)
            self.print()
        else:
            self.print('BMC does not have a FRU device.')

    # Print FRUs of other devices
    if req_id == 0:  return
    sdr_repo = get_sdr_repo(self)

    flag = False
    for sdr1 in sdr_repo[0x11]:
        fru_id = sdr1[1]
        if (fru_id == 0) or (req_id != -1 and req_id != fru_id):  continue
        if sdr1[2] & 0x80 == 0:  continue

        area_size, _ = self.intf.issue_cmd(GetFruAreaInfo, fru_id)
        if area_size == 0:  
            self.print('The area size of the FRU device with ID {0} is zero.'.format(fru_id))
            continue

        fru_name = sdr1[11:].decode('latin_1')
        self.print('{0:26}: {1} (ID {2})'.format(title, fru_name, fru_id))

        # Logical FRU
        _fru_print_one(self, fru_id, area_size)
        self.print()
        flag = True

    if not flag and req_id != -1:
        self.print('Cannot find FRU ID {0} from SDR.'.format(req_id))

def _fru_read(self, fru_id, fru_file, area_size):
    offset = 0
    with open(fru_file, 'wb') as out_file:
        while offset < area_size:
            data = self.intf.issue_cmd(ReadFru, fru_id, offset, FRU_BUF_SIZE)
            count = data[0]
            out_file.write(data[1:])
            offset += count        

    self.print('Done')

def _fru_write(self, fru_id, fru_file, area_size):
    file_size = os.stat(fru_file).st_size
    if area_size < file_size:
        raise PyCmdsExcept('ERROR: The input file size {0} is bigger than the target FRU area size {1}.'
                            .format(file_size, area_size), -1)

    self.print('Size to Write    : {0} bytes'.format(file_size))

    offset = 0
    with open(fru_file, 'rb') as in_file:
        while offset < area_size:
            chunk = in_file.read(FRU_BUF_SIZE)
            if not chunk:  break
            count, = self.intf.issue_cmd(WriteFru, fru_id, offset, chunk)
            offset += count

def _fru_read_write(self, argv):
    if len(argv) < 3:
        raise PyCmdsArgsExcept(1)

    fru_file = argv[2]
    fru_id = str2int(argv[1])
    if fru_id == -1:
        raise PyCmdsArgsExcept(3, 0, argv[1])

    area_size, _ = self.intf.issue_cmd(GetFruAreaInfo, fru_id)
    if area_size == 0:  
        raise PyCmdsExcept('The area size of the FRU device with ID {0} is zero.'.format(fru_id), -1)
    
    # print fru size
    self.print('Fru Size         : {0} bytes'.format(area_size))

    if argv[0] == 'write':
        _fru_write(self, fru_id, fru_file, area_size)
    else:
        _fru_read(self, fru_id, fru_file, area_size)

def _fru_edit_imp(self, fru_id, area_size, sec, sec_idx, sec_str):
    # get the offset of the area
    offset_c, offset_b, offset_p = _fru_get_common(self, fru_id, area_size)
    if ((sec == 'c' and offset_c == 0) or (sec == 'b' and offset_b == 0) 
        or (sec == 'p' and offset_p == 0)):
        raise PyCmdsExcept(
            'ERROR: the offset of area {0} is zero in the Common Header.'.format(
             AREA_NAMES[sec]))

    if sec == 'c': 
        offset = offset_c
        start_offset = 1
    elif sec == 'b': 
        offset = offset_b
        start_offset = 4
    elif sec == 'p': 
        offset = offset_p
        start_offset = 1
    
    # retrive the data of the area
    data = _fru_get_area(self, fru_id, offset, area_size, AREA_NAMES[sec]) 

    # seek the field to be edited
    curr_idx = 0
    get_field_len = lambda data, idx: data[idx] & 0x3f
    field_len = get_field_len(data, start_offset)
    while curr_idx < sec_idx:
        start_offset += field_len + 1
        if data[start_offset] == 0xc1:
            raise PyCmdsExcept('The index: {0} is out of range in the {1} Info Area.'.format(
                               sec_idx, AREA_NAMES[sec]), -1)

        field_len = get_field_len(data, start_offset)
        curr_idx += 1

    if len(sec_str) == field_len:
        # the length of the value doesn't change
        data_new = (data[:start_offset+1] + sec_str.encode('latin_1') 
                    + data[start_offset+1+field_len:-1])
    elif len(sec_str) < field_len:
        # the length of the value becomes shorter
        diff = field_len - len(sec_str)
        type_len = data[start_offset] - diff 
        data_new = (data[:start_offset] 
                    + type_len.to_bytes(1, byteorder='little') 
                    + sec_str.encode('latin_1')
                    + data[start_offset+1+field_len:-1]) 
        # padding
        data_new += b'\x00' * diff
    else:
        # the length of the value becomes longer
        # this needs to check the total length
        diff = len(sec_str) - field_len
        unused = len(data) - data.rfind(b'\xc1') - 2

        if diff > unused or len(sec_str) > 63:
            raise PyCmdsExcept('ERROR: Input string {0} is too long.'.sec_str)
        type_len = data[start_offset] + diff 
        data_new = (data[:start_offset] 
                    + type_len.to_bytes(1, byteorder='little') 
                    + sec_str.encode('latin_1')
                    + data[start_offset+1+field_len:-1-diff]) 

    # adding the first two bytes
    area_len = int((len(data_new) + 3) / 8)
    data_new = b'\x01' + area_len.to_bytes(1, byteorder='little') + data_new

    # calculate checksum
    data_new += checksum(data_new)

    # write fru
    total_bytes = len(data_new) 
    offset *= 8
    while total_bytes > 0:
        if total_bytes >= FRU_BUF_SIZE:
            to_write = FRU_BUF_SIZE
        else:
            to_write = total_bytes

        count, = self.intf.issue_cmd(WriteFru, fru_id, offset, data_new[:to_write])
        total_bytes -= to_write
        if not total_bytes: break
        offset += count 
        data_new = data_new[to_write:]

def _fru_edit(self, argv):
    #edit <fru id> field <section> <index> <string>
    if len(argv) < 6:
        raise PyCmdsArgsExcept(1)

    fru_id = str2int(argv[1])
    if fru_id == -1:
        raise PyCmdsArgsExcept(3, 0, argv[1])

    area_size, _ = self.intf.issue_cmd(GetFruAreaInfo, fru_id)
    if area_size == 0:  
        raise PyCmdsExcept('The area size of the FRU device with ID {0} is zero.'.format(fru_id), -1)

    if argv[2] != 'field':
        raise PyCmdsArgsExcept(3, 0, argv[2])

    sec = argv[3]
    if sec not in ('c', 'b', 'p'):        
        raise PyCmdsArgsExcept(2, 0, sec)

    sec_idx = str2int(argv[4])
    if sec_idx == -1:
        raise PyCmdsArgsExcept(3, 0, argv[4])

    sec_str = argv[5]
    
    _fru_edit_imp(self, fru_id, area_size, sec, sec_idx, sec_str)

def help_fru(self, argv=None, context=0):
    self.print('FRU Commands:')
    self.print('    print   [<fru_id>]')
    self.print('    read    <fru id> <file name>')
    self.print('    write   <fru id> <file name>')
    self.print('    edit <fru id> field <section(c,b,p)> <index> <string>')
    self.print('    help')

FRU_CMDS = {
    'print': _fru_print,
    'read': _fru_read_write,
    'write': _fru_read_write,
    'edit': _fru_edit,
    'help': help_fru, 
}

def do_fru(self, argv):
    do_command(self, argv, help_fru, FRU_CMDS)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from . _common import str2int, get_sensor_readings, print_sensor_list3, \
                     print_sensor_list4, do_command, print_sensor_type_list
from .. util.exception import PyCmdsArgsExcept

def _sensor_list(self, argv):
    opt = 3
    print_hdl = print_sensor_list3
    filter_sensor_type = 0

    if argv[0] == 'vlist':
        opt = 4
        print_hdl = print_sensor_list4

    if len(argv) > 1:
        if argv[1] == 'type':
            if len(argv) == 2:
                print_sensor_type_list(self)
                return
            else:
                filter_sensor_type = str2int(argv[2])
                if filter_sensor_type < 0:
                    raise PyCmdsArgsExcept(4, 0, argv[2])
        else:
            raise PyCmdsArgsExcept(2, 0, argv[1])

    reading_all = get_sensor_readings(self, opt, 0, filter_sensor_type, True)
    print_hdl(self, reading_all)

def _sensor_get(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1)

    sensor_num = str2int(argv[1])
    if sensor_num == -1:
        raise PyCmdsArgsExcept(3, 0, argv[1])

    opt = 4
    print_hdl = print_sensor_list4
    reading_all = get_sensor_readings(self, opt, 0, 0, True, sensor_num)
    print_hdl(self, reading_all)    

def help_sensor(self, argv=None, context=0):
    self.print('Sensor Commands:')
    self.print('    list [type [<sensor_type>]]')
    self.print('    vlist [type [<sensor_type>]]')
    self.print('    get <sensor number>')
    self.print('    help')

SENSOR_CMDS = {
    'list': _sensor_list,
    'vlist': _sensor_list,
    'get': _sensor_get,
    'help': help_sensor,    
}

def do_sensor(self, argv):
    do_command(self, argv, help_sensor, SENSOR_CMDS)
